#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2011 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""Timetable integration with SchoolTool app.
XXX: move to schooltool/app/timetable.py or similar.
"""
import zope.lifecycleevent
import zope.lifecycleevent.interfaces
from zope.proxy import sameProxiedObjects
from zope.interface import implements, implementer
from zope.component import adapts, adapter, getUtility, queryAdapter
from zope.annotation.interfaces import IAttributeAnnotatable
from zope.container.interfaces import IContainer
from zope.container.btree import BTreeContainer
from zope.intid.interfaces import IIntIds

from schooltool.app.app import InitBase, StartUpBase
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.interfaces import IApplicationPreferences
from schooltool.common import DateRange
from schooltool.course.section import InstructorsCrowd, LearnersCrowd
from schooltool.course.interfaces import ISection
from schooltool.schoolyear.subscriber import ObjectEventAdapterSubscriber
from schooltool.schoolyear.interfaces import ISchoolYear
from schooltool.securitypolicy.crowds import Crowd, ConfigurableCrowd
from schooltool.securitypolicy.crowds import AggregateCrowd
from schooltool.timetable import interfaces
from schooltool.timetable.schedule import ScheduleContainer
from schooltool.timetable.timetable import TimetableContainer

SCHEDULES_KEY = 'schooltool.timetable.schedules'
TIMETABLES_KEY = 'schooltool.timetable.timetables'


class ISchoolToolSchedules(IContainer):
    """A container of schedules."""


class SchoolToolSchedules(BTreeContainer):
    implements(ISchoolToolSchedules,
               IAttributeAnnotatable)


class TimetableInit(InitBase):
    def __call__(self):
        self.app[SCHEDULES_KEY] = SchoolToolSchedules()
        self.app[TIMETABLES_KEY] = SchoolToolSchedules()


class TimetableStartUp(StartUpBase):
    def __call__(self):
        if SCHEDULES_KEY not in self.app:
            self.app[SCHEDULES_KEY] = SchoolToolSchedules()
        if TIMETABLES_KEY not in self.app:
            self.app[TIMETABLES_KEY] = SchoolToolSchedules()


@adapter(interfaces.IHaveSchedule)
@implementer(interfaces.IScheduleContainer)
def getScheduleContainer(obj):
    int_ids = getUtility(IIntIds)
    obj_id = str(int_ids.getId(obj))
    app = ISchoolToolApplication(None)
    container = app[SCHEDULES_KEY].get(obj_id, None)
    if container is None:
        container = app[SCHEDULES_KEY][obj_id] = ScheduleContainer()
    return container


@implementer(interfaces.IHaveSchedule)
@adapter(interfaces.IScheduleContainer)
def getScheduleContainerOwner(container):
    int_ids = getUtility(IIntIds)
    obj_id = int(container.__name__)
    obj = int_ids.queryObject(obj_id)
    return obj


@implementer(interfaces.IHaveSchedule)
@adapter(interfaces.ISchedule)
def getScheduleOwner(schedule):
    container = schedule.__parent__
    return interfaces.IHaveSchedule(container, None)


@implementer(interfaces.ITimetableContainer)
@adapter(interfaces.ITimetable)
def getTimetableParent(timetable):
    return timetable.__parent__


@adapter(interfaces.IHaveTimetables)
@implementer(interfaces.ITimetableContainer)
def getTimetableContainer(obj):
    int_ids = getUtility(IIntIds)
    obj_id = str(int_ids.getId(obj))
    app = ISchoolToolApplication(None)
    container = app[TIMETABLES_KEY].get(obj_id, None)
    if container is None:
        container = app[TIMETABLES_KEY][obj_id] = TimetableContainer()
    return container


@adapter(interfaces.ITimetableContainer)
@implementer(interfaces.IHaveTimetables)
def getTimetableContainerOwner(container):
    int_ids = getUtility(IIntIds)
    obj_id = int(container.__name__)
    obj = int_ids.queryObject(obj_id)
    return obj


@implementer(interfaces.IHaveTimetables)
@adapter(interfaces.ISchedule)
def getScheduleTimetableOwner(schedule):
    container = schedule.__parent__
    return interfaces.IHaveTimetables(container, None)


def activityVocabularyFactory():
    return lambda context: interfaces.activity_types


class UpdateSelectedPeriodsSchedules(ObjectEventAdapterSubscriber):
    adapts(zope.lifecycleevent.interfaces.IObjectModifiedEvent,
           interfaces.ITimetable)

    def __call__(self):
        app = ISchoolToolApplication(None)
        # XXX: extremely nasty loop through all schedules.
        schedule_containers = app[SCHEDULES_KEY]
        for container in schedule_containers.values():
            notify_container = False
            for schedule in container.values():
                if (interfaces.ISelectedPeriodsSchedule.providedBy(schedule) and
                    sameProxiedObjects(schedule.timetable, self.object)):
                    notify_container = True
            if notify_container:
                zope.lifecycleevent.modified(container)


class RemoveRelatedSelectedPeriodsSchedules(ObjectEventAdapterSubscriber):
    adapts(zope.lifecycleevent.interfaces.IObjectRemovedEvent,
           interfaces.ITimetable)

    def __call__(self):
        app = ISchoolToolApplication(None)
        # XXX: extremely nasty loop through all schedules.
        schedule_containers = app[SCHEDULES_KEY]
        for container in schedule_containers.values():
            for schedule in container.values():
                if (interfaces.ISelectedPeriodsSchedule.providedBy(schedule) and
                    sameProxiedObjects(schedule.timetable, self.object)):
                    del container[schedule.__name__]


class SchooldaysForSchedule(object):
    adapts(interfaces.ISchedule)
    implements(interfaces.ISchooldays)

    def __init__(self, context):
        self.schedule = context

    @property
    def schoolyear(self):
        owner = interfaces.IHaveSchedule(self.schedule)
        return ISchoolYear(owner)

    def __contains__(self, date):
        for term in self.schoolyear.values():
            if date in term:
                return term.isSchoolday(date)
        return False

    def __iter__(self):
        schedule = self.schedule
        dates = DateRange(schedule.first, schedule.last)
        return self.iterDates(dates)

    def iterDates(self, dates):
        for date in dates:
            if date in self:
                yield date


class SchooldaysForTimetable(SchooldaysForSchedule):
    adapts(interfaces.ITimetable)

    @property
    def schoolyear(self):
        owner = interfaces.IHaveTimetables(self.schedule)
        return ISchoolYear(owner)


class AdaptingParentCrowdTemplate(Crowd):
    """A crowd that contains principals who are allowed to access the context."""

    adapter = None
    interface = None
    permission = ''

    def contains(self, principal):
        adapted = self.adapter(self.context)
        pcrowd = queryAdapter(adapted, self.interface, self.permission,
                              default=None)
        if pcrowd is not None:
            return pcrowd.contains(principal)
        else:
            return False


class ScheduleViewersCrowd(AdaptingParentCrowdTemplate):
    adapter = interfaces.IHaveSchedule
    interface = interfaces.IScheduleParentCrowd
    permission = "schooltool.view"


class ScheduleEditorsCrowd(AdaptingParentCrowdTemplate):
    adapter = interfaces.IHaveSchedule
    interface = interfaces.IScheduleParentCrowd
    permission = "schooltool.edit"


class SectionScheduleViewers(AggregateCrowd):
    """Crowd of those who can see the section schedule."""
    adapts(ISection)

    def crowdFactories(self):
        return [InstructorsCrowd, LearnersCrowd]


class ConfigurableScheduleEditors(ConfigurableCrowd):
    setting_key = 'instructors_can_schedule_sections'


class SectionScheduleEditors(AggregateCrowd):
    """Crowd of those who can see the section schedule."""
    adapts(ISection)

    def contains(self, principal):
        setting = ConfigurableScheduleEditors(self.context)
        if not setting.contains(principal):
            return False
        section = ISection(self.context, None)
        if section is None:
            return False
        contains = InstructorsCrowd(section).contains(principal)
        return contains

    def crowdFactories(self):
        return [ConfigurableScheduleEditors]


class SynchronizeScheduleTimezones(ObjectEventAdapterSubscriber):
    adapts(zope.lifecycleevent.interfaces.IObjectModifiedEvent,
           IApplicationPreferences)

    def __call__(self):
        prefs = self.object
        app = prefs.__parent__
        schedule_containers = app[SCHEDULES_KEY]
        for container in schedule_containers.values():
            if container.timezone != prefs.timezone:
                container.timezone = prefs.timezone
                zope.lifecycleevent.modified(container)


class SynchronizeTimetableTimezones(ObjectEventAdapterSubscriber):
    adapts(zope.lifecycleevent.interfaces.IObjectModifiedEvent,
           IApplicationPreferences)

    def __call__(self):
        prefs = self.object
        app = prefs.__parent__
        schedule_containers = app[TIMETABLES_KEY]
        for container in schedule_containers.values():
            for timetable in container.values():
                if timetable.timezone != prefs.timezone:
                    timetable.timezone = prefs.timezone
                    zope.lifecycleevent.modified(timetable)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for resources
"""
import unittest
import doctest

from zope.interface.verify import verifyObject

from schooltool.testing.util import run_unit_tests


def doctest_ResourceContainer():
    """Tests for ResourceContainer

        >>> from schooltool.resource.interfaces import IResourceContainer
        >>> from schooltool.resource.resource import ResourceContainer
        >>> c = ResourceContainer()
        >>> verifyObject(IResourceContainer, c)
        True

    Let's make sure it acts like a proper container should act

        >>> from zope.container.tests.test_btree import TestBTreeContainer
        >>> class Test(TestBTreeContainer):
        ...    def makeTestObject(self):
        ...        return ResourceContainer()
        >>> run_unit_tests(Test)
    """


def doctest_Resource():
    r"""Tests for Resource

        >>> from schooltool.resource.interfaces import IResource
        >>> from schooltool.resource.resource import Resource
        >>> resource = Resource()
        >>> verifyObject(IResource, resource)
        True

    Resources can have titles and descriptions too

        >>> blender = Resource(title='Blender', description="It's broken.")
        >>> blender.title
        'Blender'
        >>> blender.description
        "It's broken."

        >>> blender.notes
        u''

    """


def doctest_Location():
    r"""Test for Location

        >>> from schooltool.resource.interfaces import ILocation
        >>> from schooltool.resource.resource import Location
        >>> room = Location()
        >>> verifyObject(ILocation, room)
        True

    Locations have several standard peices of information associated with
    them:

        >>> room.capacity

    """


def doctest_Equipment():
    r"""Test for Equipment

        >>> from schooltool.resource.interfaces import IEquipment
        >>> from schooltool.resource.resource import Equipment
        >>> projector = Equipment()
        >>> verifyObject(IEquipment, projector)
        True

    In addition to standard attributes, equipment also has the following
    attributes:

        >>> projector.manufacturer
        u''
        >>> projector.model
        u''
        >>> projector.serialNumber
        u''
        >>> projector.purchaseDate
    """


def doctest_ResourceDemographics():
    r"""Test Resource Demographics objects and adapters

        >>> from schooltool.resource import interfaces, resource
        >>> from schooltool.basicperson import demographics

    First we need to set up a mock app and register its adapter:

        >>> from zope.component import provideAdapter
        >>> from zope.interface import implements
        >>> from schooltool.app.interfaces import ISchoolToolApplication

        >>> class MockSchoolToolApplication(dict):
        ...     implements(ISchoolToolApplication)
        >>> app = MockSchoolToolApplication()
        >>> provideAdapter(lambda context: app, (None,), ISchoolToolApplication)

    We need to do what the AppInit adapter would otherwise do:

        >>> resource.ResourceInit(app)()
        >>> resource.RESOURCE_DEMO_FIELDS_KEY in app
        True
        >>> resource.RESOURCE_DEMO_DATA_KEY in app
        True

    There's an adapter for the resource demo fields container:

        >>> provideAdapter(resource.getResourceDemographicsFields)
        >>> dfs = interfaces.IResourceDemographicsFields(app)
        >>> interfaces.IResourceDemographicsFields.providedBy(dfs)
        True
        >>> len(dfs)
        0

    We'll add some demo fields to the container, some that are limited to a
    specific resource type or types:

        >>> dfs['ID'] = demographics.TextFieldDescription("ID", "Identifier")
        >>> dfs['square_ft'] = demographics.TextFieldDescription("square_ft",
        ...     "Square Feet", limit_keys=['location'])
        >>> dfs['warranty'] = demographics.TextFieldDescription("warranty",
        ...     "Warranty", limit_keys=['equiptment'])
        >>> dfs['creation_date'] = demographics.DateFieldDescription(
        ...      "creation_date", "Creation Date",
        ...      limit_keys=['location', 'equiptment'])

    When we pass the filter_key method a key that does not
    belong  to any of the limit_keys lists, then it will only return
    those fields that have empty limit_keys lists.

        >>> [f.__name__ for f in dfs.filter_key('anything')]
        [u'ID']

    When we pass 'location', it picks up the additional fields that are for
    location type resources.

        >>> [f.__name__ for f in dfs.filter_key('location')]
        [u'ID', u'square_ft', u'creation_date']

    When we pass 'equiptment', it picks up the additional fields that are for
    equiptment type resources.

        >>> [f.__name__ for f in dfs.filter_key('equiptment')]
        [u'ID', u'warranty', u'creation_date']

    Finally there's an adapter that adapts a resource it's demo data:

        >>> provideAdapter(resource.getResourceDemographics)

    Now we will create a resource and see what we get when we adapt it to
    IDemographics:

        >>> sample = resource.Resource('Sample Resource')
        >>> sample.__name__ = 'sample'
        >>> demos = interfaces.IResourceDemographics(sample)
        >>> interfaces.IResourceDemographics.providedBy(demos)
        True
        >>> len(demos)
        0
    """

def test_suite():
    return unittest.TestSuite([
                doctest.DocTestSuite(optionflags=doctest.ELLIPSIS),
           ])

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Mixins for implementing calendars.
"""

import datetime
from pytz import utc
from zope.interface import implements
from schooltool.calendar.interfaces import IExpandedCalendarEvent


class CalendarMixin(object):
    """Mixin for implementing ICalendar methods.

    You do not have to use this mixin, however, it might make implementation
    easier, albeit potentially slower.

    A class that uses this mixin must already implement ICalendar.__iter__.

        >>> from schooltool.calendar.interfaces import ICalendar
        >>> from zope.interface import implements
        >>> class MyCalendar(CalendarMixin):
        ...     implements(ICalendar)
        ...     def __iter__(self):
        ...         return iter([])
        >>> from zope.interface.verify import verifyObject
        >>> verifyObject(ICalendar, MyCalendar())
        True

    """

    def __len__(self):
        """Return the number of events in this calendar.

            >>> class MyCalendar(CalendarMixin):
            ...     def __iter__(self):
            ...         return iter(['event' for n in range(3)])

            >>> cal = MyCalendar()
            >>> len(cal)
            3

        """
        count = 0
        for i in self:
            count += 1
        return count

    def find(self, unique_id):
        """Find a calendar event with a given UID.

        This particular implementation simply performs a linear search by
        iterating over all events and looking at their UIDs.

            >>> from schooltool.calendar.interfaces import ICalendar
            >>> from zope.interface import implements

            >>> class Event(object):
            ...     def __init__(self, uid):
            ...         self.unique_id = uid

            >>> class MyCalendar(CalendarMixin):
            ...     implements(ICalendar)
            ...     def __iter__(self):
            ...         return iter([Event(uid) for uid in 'a', 'b'])
            >>> cal = MyCalendar()

            >>> cal.find('a').unique_id
            'a'
            >>> cal.find('b').unique_id
            'b'
            >>> cal.find('c')
            Traceback (most recent call last):
              ...
            KeyError: 'c'

        """
        for event in self:
            if event.unique_id == unique_id:
                return event
        raise KeyError(unique_id)

    def expand(self, first, last):
        """Return an iterator over all expanded events in a given time period.

        See ICalendar for more details.
        """
        assert first.tzname() is not None
        assert last.tzname() is not None

        for event in self:
            for recurrence in event.expand(first, last):
                yield recurrence


class EditableCalendarMixin(object):
    """Mixin for implementing some IEditCalendar methods.

    This mixin implements `clear` by using `removeEvent`.

        >>> class Event(object):
        ...     def __init__(self, uid):
        ...         self.unique_id = uid

        >>> class SampleCalendar(EditableCalendarMixin):
        ...     def __init__(self):
        ...         self._events = {}
        ...     def __iter__(self):
        ...         return self._events.itervalues()
        ...     def addEvent(self, event):
        ...         self._events[event.unique_id] = event
        ...     def removeEvent(self, event):
        ...         del self._events[event.unique_id]

        >>> cal = SampleCalendar()
        >>> cal.addEvent(Event('a'))
        >>> cal.addEvent(Event('b'))
        >>> cal.addEvent(Event('c'))
        >>> len(list(cal))
        3

        >>> cal.clear()
        >>> list(cal)
        []

    """

    def clear(self):
        """Remove all events from the calendar."""
        for event in list(self):
            self.removeEvent(event)


class CalendarEventMixin(object):
    """Mixin for implementing ICalendarEvent comparison methods.

    Calendar events are equal iff all their attributes are equal.  We can get a
    list of those attributes easily because ICalendarEvent is a schema.

        >>> from schooltool.calendar.interfaces import ICalendarEvent
        >>> from zope.schema import getFieldNames
        >>> all_attrs = getFieldNames(ICalendarEvent)
        >>> 'unique_id' in all_attrs
        True
        >>> 'allday' in all_attrs
        True
        >>> '__eq__' not in all_attrs
        True

    We will create a bunch of Event objects that differ in exactly one
    attribute and compare them.

        >>> class Event(CalendarEventMixin):
        ...     def __init__(self, **kw):
        ...         for attr in all_attrs:
        ...             setattr(self, attr, '%s_default_value' % attr)
        ...         for attr, value in kw.items():
        ...             setattr(self, attr, value)

        >>> e1 = Event()
        >>> for attr in all_attrs:
        ...     e2 = Event()
        ...     setattr(e2, attr, 'some other value')
        ...     assert e1 != e2, 'change in %s was not noticed' % attr

    If you have two events with the same values in all ICalendarEvent
    attributes, they are equal

        >>> e1 = Event()
        >>> e2 = Event()
        >>> e1 == e2
        True

    even if they have extra attributes

        >>> e1 = Event()
        >>> e1.annotation = 'a'
        >>> e2 = Event()
        >>> e2.annotation = 'b'
        >>> e1 == e2
        True

    Events are ordered by their date and time, title and, finally, UID (to
    break any ties and provide a stable consistent ordering).

        >>> from datetime import datetime

        >>> e1 = Event(dtstart=datetime(2004, 12, 15))
        >>> e2 = Event(dtstart=datetime(2004, 12, 16))
        >>> e1 < e2
        True

        >>> e1 = Event(dtstart=datetime(2004, 12, 15), title="A")
        >>> e2 = Event(dtstart=datetime(2004, 12, 15), title="B")
        >>> e1 < e2
        True

        >>> e1 = Event(dtstart=datetime(2004, 12, 1), title="A", unique_id="X")
        >>> e2 = Event(dtstart=datetime(2004, 12, 1), title="A", unique_id="Y")
        >>> e1 < e2
        True

    """

    def __eq__(self, other):
        """Check whether two calendar events are equal."""
        return (self.unique_id, self.dtstart, self.duration, self.title,
                self.description, self.location, self.recurrence, self.allday)\
               == (other.unique_id, other.dtstart, other.duration, other.title,
                   other.description, other.location, other.recurrence,
                   other.allday)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        return (self.dtstart, self.title, self.unique_id) \
               < (other.dtstart, other.title, other.unique_id)

    def __gt__(self, other):
        return (self.dtstart, self.title, self.unique_id) \
               > (other.dtstart, other.title, other.unique_id)

    def __le__(self, other):
        return (self.dtstart, self.title, self.unique_id) \
               <= (other.dtstart, other.title, other.unique_id)

    def __ge__(self, other):
        return (self.dtstart, self.title, self.unique_id) \
               >= (other.dtstart, other.title, other.unique_id)

    def hasOccurrences(self):
        """Check whether the event has any occurrences."""
        if self.recurrence is None:
            # No recurrence rule implies one and only one occurrence.
            return True
        if self.recurrence.until is None and self.recurrence.count is None:
            # Events that repeat forever always have occurrences.  There is a
            # finite number of exceptions, so they cannon cover an infinite
            # numer of recurrences.
            return True
        try:
            self.recurrence.apply(self).next()
        except StopIteration:
            # No occurrences.
            return False
        else:
            # At least one occurrence exists.
            return True

    def expand(self, first, last):
        """Return an iterator over all expanded events in a given time period.

        See ICalendarEvent for more details.
        """
        zero = datetime.timedelta(0)
        epsilon = datetime.timedelta.resolution
        if self.recurrence is not None:
            # XXX mg: doesn't dtstart.time() already return UTC time?
            naivetime = self.dtstart.time()
            starttime = naivetime.replace(tzinfo=utc)
            # XXX mg: I have a bad feeling about taking just the date part and
            #         discarding the time and timezone.  It should at least
            #         convert the dates to UTC!
            for recdate in self.recurrence.apply(self, first.date(),
                                                 last.date()):
                dtstart = datetime.datetime.combine(recdate, starttime)
                dtend = dtstart + self.duration
                if self.duration == zero: # corner case: zero-length self
                    dtend += epsilon      # treat it as a very short self
                if dtend > first and dtstart < last:
                    yield ExpandedCalendarEvent(self, dtstart=dtstart)
        else:
            dtstart = self.dtstart
            dtend = dtstart + self.duration
            if self.duration == zero: # corner case: zero-length self
                dtend += epsilon      # treat it as a very short self
            if dtend > first and dtstart < last:
                yield self


class ExpandedCalendarEvent(CalendarEventMixin):
    """A single occurrence of a recurring calendar event.

    When creating an expanded event, you must specify the original recurrent
    event.

        >>> from schooltool.calendar.simple import SimpleCalendarEvent
        >>> from schooltool.calendar.recurrent import DailyRecurrenceRule
        >>> dtstart = datetime.datetime(2005, 2, 10, 1, 2)
        >>> duration = datetime.timedelta(hours=3)
        >>> recurrence = DailyRecurrenceRule()
        >>> original = SimpleCalendarEvent(dtstart, duration, "An event",
        ...                                description="Some description",
        ...                                unique_id="some unique id",
        ...                                location="Out in the open",
        ...                                recurrence=recurrence)

        >>> dtstart2 = datetime.datetime(2005, 2, 11, 1, 2,tzinfo=utc)
        >>> dtstart2.date()
        datetime.date(2005, 2, 11)
        >>> dtstart2.time()
        datetime.time(1, 2)
        >>> dtstart2.tzname()
        'UTC'
        >>> evt = ExpandedCalendarEvent(original, dtstart2)

        >>> from zope.interface.verify import verifyObject
        >>> verifyObject(IExpandedCalendarEvent, evt)
        True

    The start date of the event will be the specified one:

        >>> evt.dtstart.date()
        datetime.date(2005, 2, 11)
        >>> evt.dtstart.time()
        datetime.time(1, 2)
        >>> evt.dtstart.tzname()
        'UTC'

    Other attributes will be the same as in the original event:

        >>> evt.duration
        datetime.timedelta(0, 10800)

        >>> evt.title
        'An event'

        >>> evt.description
        'Some description'

        >>> evt.location
        'Out in the open'

        >>> evt.recurrence
        DailyRecurrenceRule(1, None, None, ())

    Attribute values may not be modified:

        >>> evt.dtstart = 'b0rk'
        Traceback (most recent call last):
        ...
        AttributeError: can't set attribute

        >>> evt.location = 'b0rk'
        Traceback (most recent call last):
        ...
        AttributeError: can't set attribute

    """

    implements(IExpandedCalendarEvent)

    def __init__(self, event, dtstart):
        if dtstart.tzname() not in (None, 'UTC'):
            raise ValueError, 'Can not store non UTC time info'
        self.__dict__["original"] = event
        self.__dict__["dtstart"] = dtstart.replace(tzinfo=utc)

    def __getattr__(self, name):
        return getattr(self.original, name)

    def __setattr__(self, name, value):
        raise AttributeError("can't set attribute")



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Tests for SchoolTool calendaring views.
"""

import unittest
import doctest
from pprint import pprint
from datetime import datetime, date, timedelta

from zope.component import provideAdapter
from zope.interface import implements
from zope.publisher.browser import TestRequest
from zope.app.testing import setup
from zope.annotation.interfaces import IAttributeAnnotatable

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.interfaces import IApplicationPreferences
from schooltool.calendar.utils import parse_date
from schooltool.app.cal import CalendarEvent
from schooltool.app.interfaces import ISchoolToolCalendar
from schooltool.person.person import Person
from schooltool.resource.resource import Resource
from schooltool.testing import setup as sbsetup
from schooltool.app.app import getApplicationPreferences
from schooltool.app.browser.pdfcal import (
    DailyPDFCalendarView,
    WeeklyPDFCalendarView,
    MonthlyPDFCalendarView)


class ApplicationStub(object):
    implements(ISchoolToolApplication, IAttributeAnnotatable)
    def __init__(self):
        pass


def stub_cal_class(klass, extra_calendars=[]):
    extra_calendars = extra_calendars[:]
    class StubbedCalendarView(klass):
        def getCalendars(self):
            return [self.context] + extra_calendars
        def dateTitle(self):
            return parse_date(self.request['date'])
    return StubbedCalendarView


def doctest_DailyPDFCalendarView():
    """Tests for DailyPDFCalendarView basic methods and properties.

        >>> request = TestRequest(form={'date': '2005-07-08'})
        >>> person = Person(title="Mr. Smith")
        >>> calendar = ISchoolToolCalendar(person)
        >>> view = stub_cal_class(DailyPDFCalendarView)(calendar, request)

    Daily view has a title.

        >>> print view.owner
        Mr. Smith

        >>> print view.title
        Daily calendar for Mr. Smith

    But no subtitle.

        >>> print view.subtitle
        <BLANKLINE>

    We stubbed the date retrieval mechanism to parse the date from request.
    In reality, if 'date' is absent in request, today is returned.

        >>> view.getDate()
        datetime.date(2005, 7, 8)

        >>> print view.dayTitle(view.getDate())
        2005-07-08, Friday

    """


def doctest_DailyPDFCalendarView_getCalendars(self):
    """Test for DailyPDFCalendarView.getCalendars().

    getCalendars() only delegates the task to the ICalendarProvider
    subscriber.  We will provide a stub subscriber to test the method.

        >>> class CalendarListSubscriberStub(object):
        ...     def __init__(self,context, request):
        ...         pass
        ...     def getCalendars(self):
        ...         return [('some calendar', 'color1', 'color2'),
        ...                 ('another calendar', 'color1', 'color2')]

        >>> from zope.component import provideSubscriptionAdapter
        >>> from zope.publisher.interfaces.http import IHTTPRequest
        >>> from schooltool.app.interfaces import ISchoolToolCalendar
        >>> from schooltool.app.browser.interfaces import ICalendarProvider
        >>> provideSubscriptionAdapter(CalendarListSubscriberStub,
        ...                            (ISchoolToolCalendar, IHTTPRequest),
        ...                            ICalendarProvider)

        >>> from schooltool.app.cal import Calendar
        >>> view = DailyPDFCalendarView(Calendar(None), TestRequest())

    Now, if we call the method, the output of our stub will be returned:

        >>> view.getCalendars()
        ['some calendar', 'another calendar']

    """


def doctest_DailyPDFCalendarView_getTimezone():
    """Tests for DailyPDFCalendarView.getTimezone.

    We need some extra setup here:

        >>> from schooltool.app.interfaces import ISchoolToolCalendar
        >>> setup.setUpAnnotations()

        >>> request = TestRequest(form={'date': '2005-07-08'})
        >>> person = Person(title="Mr. Smith")
        >>> calendar = ISchoolToolCalendar(person)
        >>> view = stub_cal_class(DailyPDFCalendarView)(calendar, request)
        >>> view.getTimezone()
        <UTC>

        >>> app = ISchoolToolApplication(None)
        >>> IApplicationPreferences(app).timezone = "Europe/Vilnius"

        >>> from pytz import timezone
        >>> view.getTimezone() == timezone('Europe/Vilnius')
        True

    """


def doctest_DailyPDFCalendarViewBase_tables():
    r"""Tests for DailyPDFCalendarView.tables.


        >>> calendar = ISchoolToolCalendar(Person(title="Mr. Smith"))
        >>> request = TestRequest(form={'date': '2005-07-08'})
        >>> view = stub_cal_class(DailyPDFCalendarView)(calendar, request)

    There is only a single table for the current day.

        >>> view.buildDayTable = lambda date: 'Table for %s' % date
        >>> view.tables()
        ['Table for 2005-07-08']

    """


def doctest_DailyPDFCalendarViewBase_dayEvents():
    """Event listing tests.

        >>> calendar = ISchoolToolCalendar(Person(title="Mr. Smith"))
        >>> resource = Resource()
        >>> calendar2 = ISchoolToolCalendar(resource)
        >>> request = TestRequest(form={'date': '2005-07-08'})
        >>> view = stub_cal_class(DailyPDFCalendarView)(calendar, request)
        >>> view.getCalendars = lambda: [calendar, calendar2]

    First check the simple case when the calendar is empty:

        >>> view.dayEvents(date(2005, 7, 8))
        []

    Let's add one event.

        >>> evt = CalendarEvent(datetime(2005, 7, 8, 9, 10),
        ...                     timedelta(hours=5), "evt")
        >>> calendar.addEvent(evt)

    The event should appear in the result

        >>> view.dayEvents(date(2005, 7, 8)) == [evt]
        True

        >>> view.dayEvents(date(2005, 7, 9))
        []

    We will add some events to the other calendar to test overlaying.
    If several events occur, they should be returned sorted by start time:

        >>> evt2 = CalendarEvent(datetime(2005, 7, 8, 9, 12),
        ...                      timedelta(hours=5), "evt2")
        >>> calendar2.addEvent(evt2)

    Let's add a recurring event to check expansion:

        >>> from schooltool.calendar.recurrent import DailyRecurrenceRule
        >>> evt3 = CalendarEvent(datetime(2005, 7, 5, 9, 3),
        ...                      timedelta(hours=2), "evt3",
        ...                      recurrence=DailyRecurrenceRule())
        >>> calendar2.addEvent(evt3)

        >>> result = view.dayEvents(date(2005, 7, 8))
        >>> [event.title for event in result]
        ['evt3', 'evt', 'evt2']

    All-day events always appear in front:

        >>> ad_evt = CalendarEvent(datetime(2005, 7, 8, 20, 3),
        ...                        timedelta(hours=2), "allday", allday=True)
        >>> calendar.addEvent(ad_evt)

        >>> result = view.dayEvents(date(2005, 7, 8))
        >>> [event.title for event in result]
        ['allday', 'evt3', 'evt', 'evt2']

    Booked event dupes are eliminated:

        >>> evt.bookResource(resource)
        >>> result = view.dayEvents(date(2005, 7, 8))
        >>> [event.title for event in result]
        ['allday', 'evt3', 'evt', 'evt2']

    """


def doctest_DailyPDFCalendarView_dayEvents_timezone():
    """Let's test that dayEvents handles timezones correctly.

    First' let's someone setup the user a timezone:

        >>> from schooltool.app.interfaces import ISchoolToolCalendar
        >>> setup.setUpAnnotations()
        >>> app = ISchoolToolApplication(None)
        >>> IApplicationPreferences(app).timezone = "Europe/Vilnius"

    Let's create a calendar and a view:

        >>> calendar = ISchoolToolCalendar(Person(title="Mr. Smith"))
        >>> request = TestRequest()
        >>> view = stub_cal_class(DailyPDFCalendarView)(calendar, request)

    Let's add several edge-case events to the user's calendar:

        >>> from pytz import utc
        >>> calendar.addEvent(
        ...     CalendarEvent(datetime(2005, 7, 7, 20, 0, tzinfo=utc),
        ...                   timedelta(minutes=10), '20Z'))
        >>> calendar.addEvent(
        ...     CalendarEvent(datetime(2005, 7, 7, 21, 0, tzinfo=utc),
        ...                   timedelta(minutes=10), '21Z'))
        >>> calendar.addEvent(
        ...     CalendarEvent(datetime(2005, 7, 7, 22, 0, tzinfo=utc),
        ...                   timedelta(minutes=10), '22Z'))

        >>> calendar.addEvent(
        ...     CalendarEvent(datetime(2005, 7, 8, 19, 0, tzinfo=utc),
        ...                   timedelta(minutes=10), '19Z+1d'))
        >>> calendar.addEvent(
        ...     CalendarEvent(datetime(2005, 7, 8, 20, 0, tzinfo=utc),
        ...                   timedelta(minutes=10), '20Z+1d'))
        >>> calendar.addEvent(
        ...     CalendarEvent(datetime(2005, 7, 8, 21, 0, tzinfo=utc),
        ...                   timedelta(minutes=10), '21Z+1d'))

    We should get only the events that fall into July 8 in Vilnius
    timezone:

        >>> result = view.dayEvents(date(2005, 7, 8))
        >>> [event.title for event in result]
        ['21Z', '22Z', '19Z+1d', '20Z+1d']

    """


def doctest_DailyPDFCalendarView_buildDayTable():
    r"""Tests for DailyPDFCalendarView.buildDayTable.

        >>> calendar = ISchoolToolCalendar(Person(title="Mr. Smith"))
        >>> calendar2 = ISchoolToolCalendar(Person(title='Mr. X'))
        >>> request = TestRequest(form={'date': '2005-07-08'})
        >>> ViewForTest = stub_cal_class(DailyPDFCalendarView,
        ...                             extra_calendars=[calendar2])
        >>> view = ViewForTest(calendar, request)

    Let's add an ordinary event:

        >>> from pytz import utc
        >>> evt = CalendarEvent(datetime(2005, 7, 8, 9, 10, tzinfo=utc),
        ...                     timedelta(hours=2), "Some event")
        >>> calendar.addEvent(evt)

        >>> rsrc = Resource(title='Some resource')
        >>> evt.bookResource(rsrc)

        >>> pprint(view.buildDayTable(date(2005, 7, 8)))
        {'rows': [{'description': None,
                   'location': None,
                   'resources': 'Some resource',
                   'tags': '',
                   'time': '09:10-11:10',
                   'title': 'Some event'}],
         'title': u'2005-07-08, Friday'}

        >>> calendar.clear()

    And an all-day event:

        >>> evt = CalendarEvent(datetime(2005, 7, 8, 9, 10),
        ...                     timedelta(hours=2), "Long event", allday=True)
        >>> calendar.addEvent(evt)

        >>> pprint(view.buildDayTable(date(2005, 7, 8)))
        {'rows': [{'description': None,
                   'location': None,
                   'resources': '',
                   'tags': '',
                   'time': u'all day',
                   'title': 'Long event'}],
         'title': u'2005-07-08, Friday'}

        >>> calendar.clear()

    Let's create more interesting events now. Add a recurring event with
    description and location.

        >>> from schooltool.calendar.recurrent import DailyRecurrenceRule
        >>> evt = CalendarEvent(datetime(2005, 7, 8, 9, 10),
        ...                     timedelta(hours=2), "Some recurrent event",
        ...                     location=u"\u0105 location",
        ...                     recurrence=DailyRecurrenceRule())
        >>> evt.description = u"Fun every day!"
        >>> calendar.addEvent(evt)

        >>> pprint(view.buildDayTable(date(2005, 7, 8)))
        {'rows': [{'description': u'Fun every day!',
                   'location': u'\u0105 location',
                   'resources': '',
                   'tags': u'recurrent',
                   'time': '09:10-11:10',
                   'title': 'Some recurrent event'}],
         'title': u'2005-07-08, Friday'}

        >>> calendar.clear()

    Overlaid events are also recognized.

        >>> evt2 = CalendarEvent(datetime(2005, 7, 8, 9, 10),
        ...                     timedelta(hours=2), "Recurrent event X",
        ...                     recurrence=DailyRecurrenceRule())
        >>> calendar2.addEvent(evt2)

        >>> pprint(view.buildDayTable(date(2005, 7, 8)))
        {'rows': [{'description': None,
                   'location': None,
                   'resources': '',
                   'tags': u'recurrent, from the calendar of Mr. X',
                   'time': '09:10-11:10',
                   'title': 'Recurrent event X'}],
         'title': u'2005-07-08, Friday'}

    """


def doctest_WeeklyPDFCalendarView():
    r"""Tests for WeeklyPDFCalendarView.

        >>> calendar = ISchoolToolCalendar(Person(title='John'))
        >>> request = TestRequest(form={'date': '2005-07-15'})
        >>> view = stub_cal_class(WeeklyPDFCalendarView)(calendar, request)

    Weekly view has a title, subtitle:

        >>> print view.title
        Weekly calendar for John

        >>> print view.subtitle
        Week 28 (2005-07-11 - 2005-07-17), 2005

    It builds tables for all 7 days of the week, otherwise its identical to
    the daily view.

        >>> view.buildDayTable = lambda date: 'Table for %s' % date
        >>> pprint(view.tables())
        ['Table for 2005-07-11',
         'Table for 2005-07-12',
         'Table for 2005-07-13',
         'Table for 2005-07-14',
         'Table for 2005-07-15',
         'Table for 2005-07-16',
         'Table for 2005-07-17']

    """


def doctest_MonthlyPDFCalendarView():
    r"""Tests for MonthlyPDFCalendarView.

        >>> calendar = ISchoolToolCalendar(Person(title='John'))
        >>> request = TestRequest(form={'date': '2005-02-15'})
        >>> view = stub_cal_class(MonthlyPDFCalendarView)(calendar, request)

    Monthly view has a title, subtitle:

        >>> print view.title
        Monthly calendar for John

        >>> print view.subtitle
        February, 2005

    It builds tables for every day of the month, otherwise its identical to
    the daily view.

        >>> view.buildDayTable = lambda date: 'Table for %s' % date
        >>> pprint(view.tables())
        ['Table for 2005-02-01',
         'Table for 2005-02-02',
         'Table for 2005-02-03',
         'Table for 2005-02-04',
         ...
         'Table for 2005-02-12',
         'Table for 2005-02-13',
         'Table for 2005-02-14',
         ...
         'Table for 2005-02-25',
         'Table for 2005-02-26',
         'Table for 2005-02-27',
         'Table for 2005-02-28']

    """


def pdfSetUp(test=None):
    setup.placefulSetUp()
    sbsetup.setUpCalendaring()
    app = ApplicationStub()
    provideAdapter(lambda x: app, (None,), ISchoolToolApplication)
    provideAdapter(getApplicationPreferences,
                   (ISchoolToolApplication,), IApplicationPreferences)


def pdfTearDown(test=None):
    setup.placefulTearDown()


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(doctest.DocTestSuite('schooltool.app.browser.pdfcal'))
    optionflags = (doctest.ELLIPSIS | doctest.REPORT_NDIFF
                   | doctest.NORMALIZE_WHITESPACE)
    docsuite = doctest.DocTestSuite(setUp=pdfSetUp, tearDown=pdfTearDown,
                                    optionflags=optionflags)
    suite.addTest(docsuite)
    return suite


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



# Make a package.

def registerTestSetup():
    from zope.interface import classImplements
    from schooltool.testing import registry

    def addPersonContainer(app):
        from schooltool.person import person
        app['persons'] = person.PersonContainer()
    registry.register('ApplicationContainers', addPersonContainer)

    def haveCalendar():
        from schooltool.person import person
        from schooltool.app.interfaces import IHaveCalendar
        if not IHaveCalendar.implementedBy(person.Person):
            classImplements(person.Person, IHaveCalendar)
    registry.register('CalendarComponents', haveCalendar)

registerTestSetup()
del registerTestSetup




#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2013 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Report Tests
"""
import unittest
from textwrap import dedent

from schooltool.skin.flourish.report import buildHTMLParagraphs
from schooltool.testing.util import NiceDiffsMixin


class TestBuildHTMLParagraphs(NiceDiffsMixin, unittest.TestCase):

    def test_p(self):
        snippet = 'Text'
        self.assertEqual(buildHTMLParagraphs(snippet),
                         ['Text'])
        snippet = '''<p>One paragraph</p>'''
        self.assertEqual(buildHTMLParagraphs(snippet),
                         ['One paragraph'])
        snippet = '''<p>One paragraph</p><p> </p>'''
        self.assertEqual(buildHTMLParagraphs(snippet),
                         ['One paragraph'])
        snippet = dedent('''
            <p>One paragraph</p>
            <p>Another paragraph</p>''')
        self.assertEqual(buildHTMLParagraphs(snippet),
                         ['One paragraph',
                          'Another paragraph'])

    def test_br(self):
        snippet = dedent('''
            <p>One paragraph<br/>Another line</p>''')
        self.assertEqual(buildHTMLParagraphs(snippet),
                         ['One paragraph',
                          'Another line'])
        snippet = '''<p>One paragraph<br>    <br></p>'''
        self.assertEqual(buildHTMLParagraphs(snippet),
                         ['One paragraph'])

    def test_entities(self):
        snippet = '''<p>One &amp; Two</p>'''
        self.assertEqual(buildHTMLParagraphs(snippet),
                         ['One &amp; Two'])
        snippet = '''<p>One&nbsp;&lt;&nbsp;Two</p>'''
        self.assertEqual(buildHTMLParagraphs(snippet),
                         ['One &lt; Two'])
        snippet = '''<p>One &lt; <strong>Two</strong></p>'''
        self.assertEqual(buildHTMLParagraphs(snippet),
                         ['One &lt; <strong>Two</strong>'])

    def test_bold_italic(self):
        snippet = '''<b>Bold</b> or <i>Italic</i>'''
        self.assertEqual(buildHTMLParagraphs(snippet),
                         ['<b>Bold</b> or <i>Italic</i>'])
        snippet = '''<strong>Bold</strong> or <em>Italic</em>'''
        self.assertEqual(buildHTMLParagraphs(snippet),
                         ['<strong>Bold</strong> or <em>Italic</em>'])

    def test_other_tags(self):
        snippet = '''<ul><li>One</li><li>Two</li></ul>'''
        self.assertEqual(buildHTMLParagraphs(snippet),
                         ['&lt;ul&gt;&lt;li&gt;One&lt;/li&gt;&lt;li&gt;Two&lt;/li&gt;&lt;/ul&gt;'])


if __name__ == '__main__':
    unittest.main()



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Tests for schooltool views.
"""
import unittest
import doctest

from zope.publisher.browser import TestRequest
from zope.browserpage.simpleviewclass import SimpleViewClass
from zope.app.testing import setup
from zope.traversing.browser.absoluteurl import absoluteURL
from zope.component import provideAdapter, provideUtility

from schooltool.app.browser.testing import setUp, tearDown
from schooltool.testing import setup as sbsetup


def doctest_ApplicationView():
    r"""Test for ApplicationView

    Some setup

        >>> sbsetup.setUpCalendaring()

        >>> from schooltool.app.app import getApplicationPreferences
        >>> from schooltool.app.interfaces import IApplicationPreferences
        >>> from schooltool.app.interfaces import ISchoolToolApplication

        >>> app = sbsetup.setUpSchoolToolSite()

        >>> provideAdapter(getApplicationPreferences,
        ...                (ISchoolToolApplication,), IApplicationPreferences)

    Now let's create a view

        >>> from schooltool.app.browser.app import ApplicationView
        >>> request = TestRequest()
        >>> view = ApplicationView(app, request)
        >>> view.update()

        >>> request.response.getStatus()
        302
        >>> request.response.getHeader('Location')
        'http://127.0.0.1/calendar'

    If we change a the front page preference, we are redirected
    to the login page

        >>> IApplicationPreferences(app).frontPageCalendar = False
        >>> request = TestRequest()
        >>> view = ApplicationView(app, request)
        >>> view.update()

        >>> request.response.getStatus()
        302
        >>> request.response.getHeader('Location')
        'http://127.0.0.1/auth/@@login.html'

    """


def doctest_LoginView():
    """

    Some framework setup:

        >>> setup.setUpAnnotations()

        >>> def getCalendarWithNakedObject(obj):
        ...     from schooltool.app.cal import getCalendar
        ...     from zope.security.proxy import removeSecurityProxy
        ...     return getCalendar(removeSecurityProxy(obj))

        >>> from schooltool.app.interfaces import IHaveCalendar
        >>> from schooltool.app.interfaces import ISchoolToolCalendar
        >>> provideAdapter(getCalendarWithNakedObject,
        ...                (IHaveCalendar,), ISchoolToolCalendar)

        >>> from schooltool.testing import registry
        >>> registry.setupCalendarComponents()

        >>> from schooltool.app.interfaces import ISchoolToolApplication
        >>> from schooltool.app.interfaces import IApplicationPreferences
        >>> from schooltool.app.app import getApplicationPreferences
        >>> provideAdapter(getApplicationPreferences,
        ...                (ISchoolToolApplication,), IApplicationPreferences)

    We have to set up a security checker for person objects:

        >>> from schooltool.person.person import Person
        >>> from zope.security.checker import defineChecker, Checker
        >>> defineChecker(Person, Checker({},{}))

    Suppose we have a SchoolTool app and a person:

        >>> app = sbsetup.setUpSchoolToolSite()
        >>> persons = app['persons']

        >>> frog = Person('frog')
        >>> persons[None] = frog
        >>> frog.setPassword('pond')

    We create our view:

        >>> from schooltool.app.browser.app import LoginView
        >>> request = TestRequest()
        >>> from zope.security.interfaces import IPrincipal
        >>> from zope.interface import implements
        >>> class StubPrincipal(object):
        ...     implements(IPrincipal)
        ...     title = "Some user"
        ...
        >>> request.setPrincipal(StubPrincipal())
        >>> View = SimpleViewClass('../templates/login.pt', bases=(LoginView,))
        >>> view = View(app, request)

    Render it with an empty request:

        >>> content = view()
        >>> '<h3>Please log in</h3>' in content
        True

    If we have authentication utility:

        >>> from schooltool.app.security import SchoolToolAuthenticationUtility
        >>> from zope.authentication.interfaces import IAuthentication
        >>> auth = SchoolToolAuthenticationUtility()
        >>> provideUtility(auth, IAuthentication)

        >>> from schooltool.app.security import PersonContainerAuthenticationPlugin
        >>> plugin = PersonContainerAuthenticationPlugin()
        >>> provideUtility(plugin)

        >>> auth.__parent__ = app
        >>> sbsetup.setUpSessions()

    It does not authenticate our session:

        >>> auth.authenticate(request)

    However, if we pass valid credentials, we get authenticated:

        >>> request = TestRequest(form={'username': 'frog',
        ...                             'password': 'pond',
        ...                             'LOGIN': 'Log in'})
        >>> request.setPrincipal(StubPrincipal())
        >>> view = View(app, request)
        >>> content = view()
        >>> view.error
        >>> request.response.getStatus()
        302
        >>> request.response.getHeader('Location')
        'http://127.0.0.1/persons/frog/@@logindispatch'
        >>> auth.authenticate(request)
        <schooltool.app.security.Principal object at 0x...>

    If we pass bad credentials, we get a nice error and a form.

        >>> request = TestRequest(form={'username': 'snake',
        ...                             'password': 'pw',
        ...                             'LOGIN': 'Log in'})
        >>> auth.setCredentials(request, 'frog', 'pond')
        >>> request.setPrincipal(auth.authenticate(request))
        >>> view = View(app, request)
        >>> content = view()
        >>> view.error
        u'Username or password is incorrect'
        >>> view.error in content
        True
        >>> 'Please log in' in content
        True

    The previous credentials are not lost if a new login fails:

        >>> principal = auth.authenticate(request)
        >>> principal
        <schooltool.app.security.Principal object at 0x...>
        >>> principal.id
        'sb.person.frog'

    We can specify the URL we want to go to after being authenticated:

        >>> request = TestRequest(form={'username': 'frog',
        ...                             'password': 'pond',
        ...                             'nexturl': 'http://127.0.0.1/path',
        ...                             'LOGIN': 'Log in'})
        >>> request.setPrincipal(StubPrincipal())
        >>> view = View(app, request)
        >>> content = view()
        >>> view.error
        >>> request.response.getStatus()
        302
        >>> url = absoluteURL(app, request)
        >>> request.response.getHeader('Location')
        'http://127.0.0.1/path'

    But we cannot specify a different server.
        >>> request = TestRequest(form={'username': 'frog',
        ...                             'password': 'pond',
        ...                             'nexturl': 'http://FAKE/path',
        ...                             'LOGIN': 'Log in'})
        >>> request.setPrincipal(StubPrincipal())
        >>> view = View(app, request)
        >>> content = view()
        Traceback (most recent call last):
        ...
        ValueError: Untrusted redirect to host '...' not allowed.

    """


def doctest_LogoutView():
    """
    Suppose we have a SchoolTool app and a person:

        >>> from schooltool.app.interfaces import ISchoolToolApplication
        >>> from zope.component import provideAdapter
        >>> app = sbsetup.setUpSchoolToolSite()
        >>> persons = app['persons']
        >>> provideAdapter(lambda context: app, adapts=[None],
        ...                provides=ISchoolToolApplication)

        >>> from schooltool.person.person import Person
        >>> frog = Person('frog')
        >>> persons[None] = frog
        >>> frog.setPassword('pond')

    Also, we have an authentication utility:

        >>> from schooltool.app.security import SchoolToolAuthenticationUtility
        >>> from zope.authentication.interfaces import IAuthentication
        >>> auth = SchoolToolAuthenticationUtility()
        >>> provideUtility(auth, IAuthentication)
        >>> auth.__parent__ = app
        >>> sbsetup.setUpSessions()

        >>> from schooltool.app.security import PersonContainerAuthenticationPlugin
        >>> plugin = PersonContainerAuthenticationPlugin()
        >>> provideUtility(plugin)

    We have a request in an authenticated session:

        >>> request = TestRequest()
        >>> auth.setCredentials(request, 'frog', 'pond')
        >>> request.setPrincipal(auth.authenticate(request))

    And we call the logout view:

        >>> from schooltool.app.browser.app import LogoutView
        >>> view = LogoutView(app, request)
        >>> view()

    Now, the session no longer has an authenticated user:

        >>> auth.authenticate(request)

    The user gets redirected to the front page:

        >>> request.response.getStatus()
        302
        >>> url = absoluteURL(app, request)
        >>> request.response.getHeader('Location') == url
        True


    The view also doesn't fail if the user was not logged in in the
    first place:

        >>> request = TestRequest()
        >>> view = LogoutView(app, request)
        >>> view()
        >>> auth.authenticate(request)

    """


def doctest_hasPermissions():
    r"""The Zope security machinery does not have tools to check
    whether a random principal has some permission on some object.  So
    we need to construct our own.

    Set up for local grants:

        >>> from zope.annotation.interfaces import IAnnotatable
        >>> from zope.securitypolicy.interfaces import \
        ...                         IPrincipalPermissionManager
        >>> from zope.securitypolicy.principalpermission import \
        ...                         AnnotationPrincipalPermissionManager
        >>> setup.setUpAnnotations()
        >>> setup.setUpTraversal()
        >>> provideAdapter(AnnotationPrincipalPermissionManager,
        ...                (IAnnotatable,), IPrincipalPermissionManager)

    Let's set the Zope security policy:

        >>> from zope.security.management import setSecurityPolicy
        >>> from zope.securitypolicy.zopepolicy import ZopeSecurityPolicy
        >>> old = setSecurityPolicy(ZopeSecurityPolicy)

    Suppose we have a SchoolTool object:

        >>> app = sbsetup.setUpSchoolToolSite()

    In it, we have a principal:

        >>> from schooltool.person.person import Person
        >>> app['persons']['1'] = Person('joe', title='Joe')

    He does not have neither 'super' nor 'duper' permissions on our
    schooltool app:

        >>> from schooltool.app.browser.app import hasPermissions
        >>> hasPermissions(['super', 'duper'], app, 'sb.person.joe')
        [False, False]

    However, we can add a local grant:

        >>> perms = IPrincipalPermissionManager(app)
        >>> perms.grantPermissionToPrincipal('super', 'sb.person.joe')

    And everything changes!

        >>> hasPermissions(['super', 'duper'], app, 'sb.person.joe')
        [True, False]

    The same works for subobjects:

        >>> hasPermissions(['duper', 'super'], app['persons'], 'sb.person.joe')
        [False, True]
        >>> hasPermissions(['super', 'duper'], app['persons']['joe'], 'sb.person.joe')
        [True, False]

    Also, it works gracefully for None or random objects:

        >>> hasPermissions(['super'], None, 'sb.person.joe')
        [False]
        >>> hasPermissions(['super'], object(), 'sb.person.joe')
        [False]

    """


def doctest_ApplicationPreferencesView():
    """Test for ApplicationPreferencesView.

    We need to setup a SchoolToolApplication site and build our
    ISchoolToolApplication adapter:

        >>> app = sbsetup.setUpSchoolToolSite()

        >>> from schooltool.app.browser.app import ApplicationPreferencesView
        >>> from schooltool.app.app import getApplicationPreferences
        >>> from schooltool.app.interfaces import IApplicationPreferences
        >>> from schooltool.app.interfaces import ISchoolToolApplication

        >>> setup.setUpAnnotations()
        >>> provideAdapter(getApplicationPreferences,
        ...                (ISchoolToolApplication,), IApplicationPreferences)

    Make sure we can create a view:

        >>> request = TestRequest()
        >>> view = ApplicationPreferencesView(app, request)

    Now we can setup a post and set the site title:

        >>> request = TestRequest(form={
        ...     'UPDATE_SUBMIT': 'Update',
        ...     'field.title': 'Company Calendars',
        ...     'field.name_sorting': 'first_name',
        ...     'field.dateformat': '%m/%d/%y',
        ...     'field.timeformat': '%I:%M %p',
        ...     'field.weekstart': '0',
        ...     'field.timezone': 'GMT'})
        >>> view = ApplicationPreferencesView(app, request)

        >>> view.update()

        >>> prefs = getApplicationPreferences(app)
        >>> prefs.title
        u'Company Calendars'

        >>> prefs.name_sorting
        'first_name'

        >>> prefs.dateformat
        '%m/%d/%y'

        >>> prefs.timeformat
        '%I:%M %p'

        >>> prefs.weekstart
        0

        >>> prefs.timezone
        'GMT'

    """


def doctest_RelationshipViewBase():
    """Test for RelationshipViewBase.

    Let's create the view first:

        >>> from schooltool.app.browser.app import RelationshipViewBase
        >>> from zope.publisher.browser import TestRequest
        >>> request = TestRequest()
        >>> view = RelationshipViewBase(None, request)

    Update method of our view should process the form and perform
    necessary actions, like add items to the collections and most
    importantly to set up available and selected item tables:

        >>> from pprint import pprint
        >>> def FakeTableSetUp(**kwargs):
        ...     print "Setting up table:"
        ...     pprint(kwargs)
        >>> view.createTableFormatter = FakeTableSetUp
        >>> view.getAvailableItems = lambda: "<Available Items>"
        >>> view.getSelectedItems = lambda: "<Selected Items>"
        >>> view.update()
        Setting up table: {'ommit': '<Selected Items>',
                           'prefix': 'add_item'}
        Setting up table: {'batch_size': 0,
                           'filter': <function <lambda> at ...>,
                           'items': '<Selected Items>',
                           'prefix': 'remove_item'}

    """


def doctest_ContentTitle():
    """Tests for ContentTitle.

        >>> from schooltool.app.browser.app import ContentTitle

        >>> class ContentStub(object):
        ...     title='Foo'

        >>> provider = ContentTitle(ContentStub(), TestRequest(), None)
        >>> print provider()
        Foo

    """


def doctest_ContentLink():
    """Tests for ContentLink.

        >>> from schooltool.app.browser.app import ContentLink

        >>> class ContentStub(object):
        ...     title='Foo'

        >>> app = sbsetup.createSchoolToolApplication()
        >>> app['content'] = ContentStub()

        >>> provider = ContentLink(app['content'], TestRequest(), None)
        >>> print provider()
        <a href="http://127.0.0.1/content">Foo</a>

    """


def doctest_ContentLabel():
    """Tests for ContentLabel.

        >>> from schooltool.app.browser.app import ContentLabel

        >>> class ContentStub(object):
        ...     title='Foo'

        >>> app = sbsetup.createSchoolToolApplication()
        >>> app['content'] = ContentStub()

        >>> provider = ContentLabel(app['content'], TestRequest(), None)
        >>> print provider()
        <a href="http://127.0.0.1/content">Foo</a>

        >>> app['content'].label = 'This is the Foo!'
        >>> print provider()
        <a href="http://127.0.0.1/content">This is the Foo!</a>

    """


def test_suite():
    optionflags = (doctest.ELLIPSIS | doctest.REPORT_NDIFF
                   | doctest.REPORT_ONLY_FIRST_FAILURE
                   | doctest.NORMALIZE_WHITESPACE)
    suite = unittest.TestSuite()
    suite.addTest(doctest.DocTestSuite(setUp=setUp, tearDown=tearDown,
                                       optionflags=optionflags))
    return suite


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2011 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Flourish breadcrumbs.
"""

import zope.schema
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.component import queryMultiAdapter
from zope.interface import implements, Interface, Attribute
from zope.security import checkPermission
from zope.traversing.api import getParent
from zope.traversing.interfaces import IContainmentRoot
from zope.traversing.browser.absoluteurl import absoluteURL

from schooltool.skin.flourish.content import ContentProvider
from schooltool.skin.flourish import interfaces


class IBreadcrumbInfo(Interface):

    title = zope.schema.TextLine(
        title=u"Title",
        description=u"""
        The name of the breadcrumb.
        If there is not title, the breadcrumb should not be displayed.
        """,
        required=False)

    link = zope.schema.URI(
        title=u"URL",
        description=u"URL of the breadcrumb.",
        required=False)


class IBreadcrumb(IBreadcrumbInfo, interfaces.IContentProvider):

    crumb_parent = Attribute(
        u"Suggested parent context for breadcrumb lookup.")

    follow_crumb = Attribute(
        u"The next crumb to follow.")

    breadcrumbs = Attribute(
        u"List of breadcrumbs in parent-to-this order.")


class BreadCrumbInfo(object):

    def __init__(self, name, url):
        self.name = name
        self.url = url

    def __repr__(self):
        return '<%s %r>' % (self.__class__.__name__, self.name)


class Breadcrumbs(ContentProvider):
    implements(IBreadcrumb)

    template = ViewPageTemplateFile('templates/breadcrumbs.pt')

    permission = "schooltool.view"

    @property
    def title(self):
        title_content = queryMultiAdapter(
            (self.context, self.request, self.view),
            interfaces.IContentProvider, 'title')
        if title_content is None:
            return None
        title = getattr(title_content, 'title', None)
        return title

    link = None

    def checkPermission(self):
        permission = self.permission
        if permission:
            if not checkPermission(permission, self.context):
                return False
        return True

    @property
    def url(self):
        if not self.checkPermission():
            return None
        context = self.context
        url = absoluteURL(context, self.request)
        link = self.link
        if link:
            url = '%s/%s' % (url, link)
        return url

    @property
    def crumb_parent(self):
        if getattr(self.context, '__parent__', None) is None:
            return None
        if IContainmentRoot.providedBy(self.context):
            return None
        return self.context.__parent__

    @property
    def follow_crumb(self):
        parent = self.crumb_parent
        if parent is None:
            return None
        breadcrumb = queryMultiAdapter(
            (parent, self.request, self.view),
            IBreadcrumb,
            name="",
            default=None)
        return breadcrumb

    @property
    def breadcrumbs(self):
        """List of breadcrumbs.
        root > child > child > this.
        """
        breadcrumbs = [BreadCrumbInfo(self.title, self.url)]
        follow = self.follow_crumb
        if follow is not None:
            breadcrumbs = follow.breadcrumbs + breadcrumbs
        return [b for b in breadcrumbs if b.name]

    def render(self, *args, **kw):
        return self.template(*args, **kw)


class PageBreadcrumbs(Breadcrumbs):
    implements(IBreadcrumb)

    @property
    def title(self):
        title = getattr(self.view, 'subtitle', None)
        title = title or getattr(self.view, 'title', None)
        return title

    url = None

    @property
    def crumb_parent(self):
        return self.context


class TitleBreadcrumb(Breadcrumbs):

    @property
    def title(self):
        return self.context.title



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2006 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool security policy crowds.
"""

from zope.interface import implements
from zope.security.proxy import removeSecurityProxy
from zope.component import queryAdapter, queryMultiAdapter, queryUtility
from zope.component import getGlobalSiteManager
from zope.container.contained import Contained
from zope.container.btree import BTreeContainer


from schooltool.securitypolicy.interfaces import (
    ICrowdsUtility,
    IDescriptionUtility,
    ICrowd,
    IAccessControlCustomisations,
    ICrowdToDescribe,
    IDescription,
    IDescriptionGroup,
    IGroupAction,
    ICrowdDescription)

from schooltool.common import SchoolToolMessage as _


class CrowdNotRegistered(Exception):
    pass


class CrowdsUtility(object):
    implements(ICrowdsUtility)

    def __init__(self):
        self.factories = {}
        self.crowds = {}

    def getCrowdNames(self, permission, interface):
        return self.crowds.get((permission, interface), [])

    def getFactory(self, crowd_name):
        if crowd_name not in self.factories:
            raise CrowdNotRegistered(crowd_name)
        return self.factories[crowd_name]

    def getFactories(self, permission, interface):
        names = self.getCrowdNames(permission, interface)
        return [self.getFactory(name) for name in names]


def getCrowdsUtility():
    """Helper - returns crowds utility and registers new one if missing."""
    utility = queryUtility(ICrowdsUtility)
    if not utility:
        utility = CrowdsUtility()
        getGlobalSiteManager().registerUtility(utility, ICrowdsUtility)
    return utility


class DescriptionUtility(object):
    implements(IDescriptionUtility)

    def __init__(self):
        self.groups = BTreeContainer()
        self.actions_by_group = BTreeContainer()


def getDescriptionUtility():
    """Helper - returns crowd description utility and registers
    a new one if missing.
    """
    utility = queryUtility(IDescriptionUtility)
    if not utility:
        utility = DescriptionUtility()
        getGlobalSiteManager().registerUtility(utility, IDescriptionUtility)
    return utility


class Description(Contained):
    implements(IDescription)

    title = u''
    description = u''

    def __repr__(self):
        return '<%s.%s (%r)>' % (self.__class__.__module__,
                              self.__class__.__name__,
                              self.title)


class DescriptionGroup(Description):
    implements(IDescriptionGroup)


class GroupAction(Description):
    implements(IGroupAction)

    interface = None
    permission = None

    @property
    def group(self):
        if self.__parent__ is None:
            return None
        group_name = self.__parent__.__name__
        util = getDescriptionUtility()
        if group_name not in util.groups:
            return None
        return util.groups[group_name]


class Crowd(object):
    """An abstract base class for crowds."""

    implements(ICrowd)

    title = u''
    description = u''

    def __init__(self, context):
        # As crowds are used in our security policy we have to trust
        # them
        self.context = removeSecurityProxy(context)

    def contains(self, principal):
        raise NotImplementedError()

    def __str__(self):
        return "<%s>" % self.__class__.__name__

    __repr__ = __str__


class AggregateCrowd(Crowd):
    """A base class for aggregating crowds.

    Override crowdFactories to specify which crowds to aggregate.
    """

    def contains(self, principal):
        for crowdcls in self.crowdFactories():
            crowd = crowdcls(self.context)
            if crowd.contains(principal):
                return True
        return False

    def crowdFactories(self):
        raise NotImplementedError("override this in subclasses")

    def __str__(self):
        crowds = [factory(self.context) for factory in self.crowdFactories()]
        return "<AggregateCrowd crowds=%s>" % crowds

    __repr__ = __str__


class ConfigurableCrowd(Crowd):
    """A base class for calendar parent crowds.

    You only need to override `setting_key` which indicates the key
    of the corresponding security setting.
    """

    setting_key = None # override in subclasses

    @property
    def settings(self):
        # XXX: Avoid a circular import.  Very naughty.
        from schooltool.app.interfaces import ISchoolToolApplication
        app = ISchoolToolApplication(None)
        return IAccessControlCustomisations(app)

    def contains(self, principal):
        """Return the value of the related setting (True or False)."""
        return self.settings.get(self.setting_key)


class EverybodyCrowd(Crowd):
    """A crowd that contains absolutely everybody."""

    title = _(u'Everybody')
    description = _(u'Everybody, including users that are not logged in.')

    def contains(self, principal):
        return True


class OwnerCrowd(Crowd):
    """Crowd of owners.

    The crowd tries to adapt the context to IPerson.  If adaptation succeeds,
    it compares the obtained person with the current principal.
    """

    def contains(self, principal):
        from schooltool.person.interfaces import IPerson
        from schooltool.app.browser import same # XXX
        person = IPerson(principal, None)
        owner = IPerson(self.context, None)
        return person is not None and same(person, owner)


class _GroupCrowd(Crowd):
    """A base class for crowds determined by the principal's groups."""

    group = None # override this

    def contains(self, principal):
        return self.group in principal.groups


class AuthenticatedCrowd(_GroupCrowd):
    """Authenticated users."""

    title = _(u'Authenticated user')
    description = _(u'All logged in users.')

    group = 'zope.Authenticated'


class AnonymousCrowd(_GroupCrowd):
    """Anonymous users."""

    title = _(u'Anybody')
    description = _(u'Users who did not log in')

    group = 'zope.Anybody'


class SuperUserCrowd(Crowd):

    title = _(u'Super user')
    description = _(u'The super user - owner of this SchoolTool application.')

    def contains(self, principal):
        from schooltool.app.browser import same # XXX
        from schooltool.app.interfaces import ISchoolToolApplication
        from schooltool.person.interfaces import IPerson
        app = ISchoolToolApplication(None)
        persons = app['persons']
        person = IPerson(principal, None)
        return person is not None and same(person, persons.super_user)


class ManagerGroupCrowd(_GroupCrowd):
    group = 'sb.group.manager'


class ManagersCrowd(Crowd):

    def contains(self, principal):
        managers_group = ManagerGroupCrowd(self.context)
        super_users = SuperUserCrowd(self.context)
        return super_users.contains(principal) or managers_group.contains(principal)


class AdministratorsCrowd(_GroupCrowd):
    group = 'sb.group.administrators'

class ClerksCrowd(_GroupCrowd):
    group = 'sb.group.clerks'

class TeachersCrowd(_GroupCrowd):
    group = 'sb.group.teachers'

class StudentsCrowd(_GroupCrowd):
    group = 'sb.group.students'


class AdministrationCrowd(AggregateCrowd):

    def crowdFactories(self):
        return [AdministratorsCrowd, ClerksCrowd, ManagersCrowd]


class ParentCrowdTemplate(Crowd):
    """A crowd that contains principals who are allowed to access the context."""

    interface = None
    permission = ''

    def contains(self, principal):
        parent = self.context.__parent__
        pcrowd = queryAdapter(parent, self.interface, self.permission,
                              default=None)
        if pcrowd is not None:
            return pcrowd.contains(principal)
        else:
            return False


def ParentCrowd(interface, permission):
    return type('ParentCrowd',
                (ParentCrowdTemplate,),
                {'interface': interface,
                 'permission': permission})


class CrowdDescription(Description):
    implements(ICrowdDescription)

    def __init__(self, crowd, action, group):
        self.crowd, self.action, self.group = crowd, action, group


class DefaultCrowdDescription(Description):
    implements(ICrowdDescription)

    def __init__(self, crowd, action, group):
        self.crowd, self.action, self.group = crowd, action, group
        if hasattr(self.crowd, 'title'):
            self.title = self.crowd.title
        if hasattr(self.crowd, 'description'):
            self.description = self.crowd.description


class AggregateCrowdDescription(CrowdDescription):

    @property
    def description(self):
        return u', '.join([d.title for d in self.getDescriptions()
                           if d.title])

    def getFactories(self):
        return self.crowd.crowdFactories()

    def getDescriptions(self):
        factories = self.getFactories()
        descriptions =[
            getCrowdDescription(factory(None), self.action, self.group)
            for factory in factories]
        return filter(None, descriptions)


def getCrowdDescription(crowd, action, group):
    crowd_to_describe = queryMultiAdapter(
        (crowd, action, group),
        ICrowdToDescribe,
        default=None)
    if crowd_to_describe is None:
        return None
    description = queryMultiAdapter(
        (crowd_to_describe, action, group),
        ICrowdDescription,
        default=None)

    return description


def defaultCrowdToDescribe(crowd, action, group):
    return crowd


def collectCrowdDescriptions(action):
    crowds = getCrowdsUtility()
    # Mimic schooltool.securitypolicy.SchoolToolSecurityPolicy behaviour
    crowd_names = crowds.getCrowdNames(action.permission,
                                       None)
    if not crowd_names:
        crowd_names = crowds.getCrowdNames(action.permission,
                                           action.interface)

    descriptions = getDescriptionUtility()
    group = action.group

    descriptions = []
    for name in sorted(crowd_names):
        factory = crowds.getFactory(name)
        crowd = factory(None)
        descriptions.append(getCrowdDescription(crowd, action, group))
    return descriptions


def inCrowd(principal, name, context=None):
    crowds = getCrowdsUtility()
    try:
        factory = crowds.getFactory(name)
    except CrowdNotRegistered:
        return False
    crowd = factory(context)
    return crowd.contains(principal)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Group interfaces
"""

from zope.interface import Interface, Attribute
from zope.schema import TextLine, Text
from zope.container.interfaces import IContainer, IContained
from zope.container.constraints import contains, containers

from schooltool.common import SchoolToolMessage as _

class IGroupMember(Interface):
    """An object that knows the groups it is a member of."""

    groups = Attribute("""Groups (see IRelationshipProperty)""")


class IBaseGroup(Interface):
    """Group."""

    title = TextLine(
        title=_("Title"))

    description = Text(
        title=_("Description"),
        required=False)

    members = Attribute(
        """Members of the group (see IRelationshipProperty)""")


class IGroup(IBaseGroup):
    """Group."""


class IGroupContainer(IContainer):
    """Container of groups."""

    contains(IGroup)


class IGroupContained(IGroup, IContained):
    """Group contained in an IGroupContainer."""

    containers(IGroupContainer)


class IGroupContainerContainer(IContainer):
    """Container of group containers."""

    contains(IGroupContainer)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Schooltool PDF support.
"""
from reportlab.pdfbase import pdfmetrics

enabled = False

def isEnabled():
    return enabled


# ------------------
# Font configuration
# ------------------

def registerTTFont(fontname, filename):
    """Register a TrueType font with ReportLab.

    Clears up the incorrect straight-through mappings that ReportLab 1.19
    unhelpfully gives us.
    """
    from reportlab.pdfbase.ttfonts import TTFont
    import reportlab.lib.fonts

    pdfmetrics.registerFont(TTFont(fontname, filename))
    # For some reason pdfmetrics.registerFont for TrueType fonts explicitly
    # calls addMapping with incorrect straight-through mappings, at least in
    # reportlab version 1.19.  We thus need to stick our dirty fingers in
    # reportlab's internal data structures and undo those changes so that we
    # can call addMapping with correct values.
    key = fontname.lower()
    del reportlab.lib.fonts._tt2ps_map[key, 0, 0]
    del reportlab.lib.fonts._tt2ps_map[key, 0, 1]
    del reportlab.lib.fonts._tt2ps_map[key, 1, 0]
    del reportlab.lib.fonts._tt2ps_map[key, 1, 1]
    del reportlab.lib.fonts._ps2tt_map[key]


# 'Arial' is predefined in ReportLab, so we use 'Arial_Normal'

font_map = {
    'Arial_Normal': 'LiberationSans-Regular.ttf',
    'Arial_Bold': 'LiberationSans-Bold.ttf',
    'Arial_Italic': 'LiberationSans-Italic.ttf',
    'Arial_Bold_Italic': 'LiberationSans-BoldItalic.ttf',
    'Times_New_Roman': 'LiberationSerif-Regular.ttf',
    'Times_New_Roman_Bold': 'LiberationSerif-Bold.ttf',
    'Times_New_Roman_Italic': 'LiberationSerif-Italic.ttf',
    'Times_New_Roman_Bold_Italic': 'LiberationSerif-BoldItalic.ttf',
    'Ubuntu_Regular': 'Ubuntu-R.ttf',
    'Ubuntu_Bold': 'Ubuntu-B.ttf',
    'Ubuntu_Italic': 'Ubuntu-RI.ttf',
    'Ubuntu_Bold_Italic': 'Ubuntu-BI.ttf',
    }

font_directories = ()


def setUpFonts(directories=()):
    """Set up ReportGen to use Liberation and Ubuntu Fonts."""
    import reportlab.rl_config
    from reportlab.lib.fonts import addMapping

    global font_directories
    if not directories:
        directories = font_directories
    else:
        font_directories = directories

    ttfpath = reportlab.rl_config.TTFSearchPath
    for fontdir in directories:
        if fontdir not in ttfpath:
            ttfpath.append(fontdir)

    reportlab.rl_config.warnOnMissingFontGlyphs = 0

    for font_name, font_file in font_map.items():
        registerTTFont(font_name, font_file)

    addMapping('Arial_Normal', 0, 0, 'Arial_Normal')
    addMapping('Arial_Normal', 0, 1, 'Arial_Italic')
    addMapping('Arial_Normal', 1, 0, 'Arial_Bold')
    addMapping('Arial_Normal', 1, 1, 'Arial_Bold_Italic')

    addMapping('Times_New_Roman', 0, 0, 'Times_New_Roman')
    addMapping('Times_New_Roman', 0, 1, 'Times_New_Roman_Italic')
    addMapping('Times_New_Roman', 1, 0, 'Times_New_Roman_Bold')
    addMapping('Times_New_Roman', 1, 1, 'Times_New_Roman_Bold_Italic')

    # XXX: Rename to Ubuntu
    addMapping('Ubuntu_Regular', 0, 0, 'Ubuntu_Regular')
    addMapping('Ubuntu_Regular', 0, 1, 'Ubuntu_Italic')
    addMapping('Ubuntu_Regular', 1, 0, 'Ubuntu_Bold')
    addMapping('Ubuntu_Regular', 1, 1, 'Ubuntu_Bold_Italic')

    global enabled
    enabled = True


# z3c.rml >= 1.1.0 calls _reset when processing each document
# We need to register our fonts again
from reportlab.rl_config import register_reset
register_reset(setUpFonts)
del register_reset



import stesting

def registerTestSetup():
    from zope.interface import classImplements
    from schooltool.testing import registry

    def haveCalendar():
        from schooltool.group import group
        from schooltool.app.interfaces import IHaveCalendar
        if not IHaveCalendar.implementedBy(group.Group):
            classImplements(group.Group, IHaveCalendar)
    registry.register('CalendarComponents', haveCalendar)

registerTestSetup()
del registerTestSetup



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool skin containers
"""
import urllib

from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.component import queryMultiAdapter
from zope.publisher.browser import BrowserView
from zope.security.checker import canAccess
from zope.traversing.browser.absoluteurl import absoluteURL

from schooltool.table.batch import IterableBatch
from schooltool.table.table import DependableCheckboxColumn
from schooltool.table.table import url_cell_formatter
from schooltool.table.interfaces import ITableFormatter


class ContainerView(BrowserView):
    """A base view for all containers.

    Subclasses must provide the following attributes that are used in the
    page template:

        `index_title` -- Title of the index page.

    """

    @property
    def container(self):
        return self.context

    def update(self):
        if 'SEARCH' in self.request and 'CLEAR_SEARCH' not in self.request:
            searchstr = self.request['SEARCH'].lower()
            results = [item for item in self.container.values()
                       if searchstr in item.title.lower()]
            search_string = self.request['SEARCH'].encode('utf-8')
            extra_url = "&SEARCH=%s" % urllib.quote_plus(search_string)
        else:
            self.request.form['SEARCH'] = ''
            results = self.container.values()
            extra_url = ""

        self.batch = IterableBatch(results, self.request, sort_by='title',
                                   extra_url=extra_url)

    @property
    def canModify(self):
        return canAccess(self.container, '__delitem__')


class ContainerDeleteView(BrowserView):
    """A view for deleting items from container."""

    @property
    def container(self):
        return self.context

    def listIdsForDeletion(self):
        return [key for key in self.container
                if "delete.%s" % key in self.request]

    def _listItemsForDeletion(self):
        return [self.container[key] for key in self.listIdsForDeletion()]

    itemsToDelete = property(_listItemsForDeletion)

    def update(self):
        if 'CONFIRM' in self.request:
            for key in self.listIdsForDeletion():
                del self.container[key]
            self.request.response.redirect(self.nextURL())
        elif 'CANCEL' in self.request:
            self.request.response.redirect(self.nextURL())

    def nextURL(self):
        return absoluteURL(self.container, self.request)


class TableContainerView(BrowserView):
    """A base view for containers that use zc.table to display items.

    Subclasses must provide the following attributes that are used in the
    page template:

        `index_title` -- Title of the index page.

    """

    template = ViewPageTemplateFile('templates/table_container.pt')
    delete_template = ViewPageTemplateFile('templates/container_delete.pt')

    def __init__(self, context, request):
        self.request = request
        self.context = context
        self.table = queryMultiAdapter((context, request), ITableFormatter)

    def setUpTableFormatter(self, formatter):
        columns_before = []
        if self.canModify():
            columns_before = [DependableCheckboxColumn(prefix="delete",
                                                       name='delete_checkbox',
                                                       title=u'')]
        formatter.setUp(formatters=[url_cell_formatter],
                        columns_before=columns_before)

    def __call__(self):
        if 'DELETE' in self.request:
            return self.delete_template()

        self.setUpTableFormatter(self.table)
        # XXX update should be in here but as the container_delete
        # template is shared with the ContainerDeleteView and update
        # is called in the template we are not doing it here
        return self.template()

    def update(self):
        if 'CONFIRM' in self.request:
            for key in self.listIdsForDeletion():
                del self.context[key]
        # XXX: Update table formatter.  It should be set up in __call__, after
        #      update(), if rendering the correct template.  But as update is
        #      called from template for some genuine reason, I'll just stick
        #      it here for now.  I hope to find some time to fix this later on.
        self.setUpTableFormatter(self.table)

    def canModify(self):
        return canAccess(self.context, '__delitem__')

    def listIdsForDeletion(self):
        return [key for key in self.context
                if "delete.%s" % key in self.request]

    def _listItemsForDeletion(self):
        return [self.context[key] for key in self.listIdsForDeletion()]

    itemsToDelete = property(_listItemsForDeletion)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool application views.
"""
import urllib

from zope.cachedescriptors.property import Lazy
from ZODB.FileStorage.FileStorage import FileStorageError
from ZODB.interfaces import IDatabase
from zope.location.location import LocationProxy
from zope.interface import implementer
from zope.interface import implements
from zope.interface import Interface
from zope.security.interfaces import IParticipation
from zope.security.management import getSecurityPolicy
from zope.security.proxy import removeSecurityProxy
from zope.app.form.utility import getWidgetsData, setUpWidgets
from zope.app.form.browser.add import AddView
from zope.app.form.browser.editview import EditView
from zope.app.form.interfaces import IInputWidget
from zope.app.form.interfaces import WidgetsError
from zope.app.applicationcontrol.browser.runtimeinfo import RuntimeInfoView
from zope.app.applicationcontrol.browser.zodbcontrol import ZODBControlView
from zope.app.applicationcontrol.interfaces import IApplicationControl
from zope.app.applicationcontrol.interfaces import IRuntimeInfo
from zope.authentication.interfaces import IUnauthenticatedPrincipal
from zope.error.interfaces import IErrorReportingUtility
from zope.i18n import translate
from zope.publisher.browser import BrowserView
from zope.component import getMultiAdapter, queryMultiAdapter
from zope.component import getUtility
from zope.component import adapter, adapts
from zope.authentication.interfaces import IAuthentication
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.publisher.browser import BrowserPage
from zope.traversing.browser.absoluteurl import absoluteURL
from zope.schema import Int, Bool, Tuple, Choice
from z3c.form import form, field, button
from z3c.form.browser.checkbox import CheckBoxFieldWidget
from z3c.form.interfaces import DISPLAY_MODE
from zc.table.table import FormFullFormatter

from schooltool.calendar.icalendar import convert_calendar_to_ical
from schooltool.app.browser.interfaces import IManageMenuViewletManager
from schooltool.app.interfaces import ISchoolToolAuthenticationPlugin
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.interfaces import IApplicationPreferences
from schooltool.app.interfaces import IApplicationTabs
from schooltool.app.interfaces import ISchoolToolCalendar
from schooltool.app.interfaces import IAsset
from schooltool.app.utils import vocabulary
from schooltool.skin.flourish.content import ContentProvider
from schooltool.common.inlinept import InlineViewPageTemplate
from schooltool.person.interfaces import IPerson
from schooltool import table
from schooltool.table.table import CheckboxColumn
from schooltool.table.table import label_cell_formatter_factory
from schooltool.table.table import ImageInputColumn
from schooltool.table.interfaces import ITableFormatter
from schooltool.securitypolicy.crowds import inCrowd
from schooltool.skin.skin import OrderedViewletManager
from schooltool.skin import flourish
from schooltool.skin.flourish.form import Form
from schooltool.skin.flourish.form import Dialog
from schooltool.schoolyear.interfaces import ISchoolYearContainer

from schooltool.common import SchoolToolMessage as _


class ApplicationView(BrowserView):
    """A view for the main application."""

    def update(self):
        prefs = IApplicationPreferences(ISchoolToolApplication(None))
        if prefs.frontPageCalendar:
            url = absoluteURL(ISchoolToolCalendar(self.context),
                              self.request)
        else:
            url = absoluteURL(self.context,
                              self.request) + '/auth/@@login.html'
        self.request.response.redirect(url)



class BaseAddView(AddView):
    """Common functionality for adding groups and resources"""

    def nextURL(self):
        return absoluteURL(self.context.context, self.request)

    def update(self):
        if 'CANCEL' in self.request:
            self.request.response.redirect(self.nextURL())
        else:
            return AddView.update(self)


class BaseEditView(EditView):
    """An edit view for resources and groups"""

    def update(self):
        if 'CANCEL' in self.request:
            url = absoluteURL(self.context, self.request)
            self.request.response.redirect(url)
        else:
            status = EditView.update(self)
            if 'UPDATE_SUBMIT' in self.request and not self.errors:
                url = absoluteURL(self.context, self.request)
                self.request.response.redirect(url)
            return status


class RelationshipViewBase(BrowserView):
    """A base class for views that add/remove members from a relationship."""

    __call__ = ViewPageTemplateFile('templates/edit_relationships.pt')

    title = None
    current_title = None
    available_title = None

    def add(self, item):
        """Add an item to the list of selected items."""
        # Only those who can edit this section will see the view so it
        # is safe to remove the security proxy here
        collection = removeSecurityProxy(self.getCollection())
        collection.add(item)

    def remove(self, item):
        """Remove an item from selected items."""
        # Only those who can edit this section will see the view so it
        # is safe to remove the security proxy here
        collection = removeSecurityProxy(self.getCollection())
        collection.remove(item)

    def getCollection(self):
        """Return the backend storage for related objects."""
        raise NotImplementedError("Subclasses should override this method.")

    def getSelectedItems(self):
        """Return a sequence of items that are already selected."""
        return self.getCollection()

    def getAvailableItems(self):
        """Return a sequence of items that can be selected."""
        container = self.getAvailableItemsContainer()
        selected_items = set(self.getSelectedItems())
        return [p for p in container.values()
                if p not in selected_items]

    def getAvailableItemsContainer(self):
        """Returns the backend storage for available items."""
        raise NotImplementedError("Subclasses should override this method.")

    def createTableFormatter(self, **kwargs):
        prefix = kwargs['prefix']
        container = self.getAvailableItemsContainer()
        formatter = getMultiAdapter((container, self.request),
                                    ITableFormatter)
        columns_before = [CheckboxColumn(prefix=prefix, title="")]
        formatters = [label_cell_formatter_factory(prefix)]
        formatter.setUp(formatters=formatters,
                        columns_before=columns_before,
                        **kwargs)
        return formatter

    def getOmmitedItems(self):
        return self.getSelectedItems()

    def setUpTables(self):
        self.available_table = self.createTableFormatter(
            ommit=self.getOmmitedItems(),
            prefix="add_item")

        self.selected_table = self.createTableFormatter(
            filter=lambda l: l,
            items=self.getSelectedItems(),
            prefix="remove_item",
            batch_size=0)

    def update(self):
        context_url = absoluteURL(self.context, self.request)

        if 'ADD_ITEMS' in self.request:
            for item in self.getAvailableItems():
                if 'add_item.' + item.__name__ in self.request:
                    self.add(removeSecurityProxy(item))
        elif 'REMOVE_ITEMS' in self.request:
            for item in self.getSelectedItems():
                if 'remove_item.' + item.__name__ in self.request:
                    self.remove(removeSecurityProxy(item))
        elif 'CANCEL' in self.request:
            self.request.response.redirect(context_url)

        self.setUpTables()


class CSSFormatter(table.table.SortUIHeaderMixin, FormFullFormatter):

    def renderHeaders(self):
        result = []
        old_css_class = self.cssClasses.get('th')
        for col in self.visible_columns:
            self.cssClasses['th'] = col.name.replace('_', '-')
            result.append(self.renderHeader(col))
        self.cssClasses['th'] = old_css_class
        return ''.join(result)


class RelationshipButton(ImageInputColumn):

    def params(self, item, formatter):
        params = ImageInputColumn.params(self, item, formatter)
        params['tokens_name'] = ".".join(
            filter(None, ["displayed", self.prefix, "tokens"]))
        params['tokens_value'] = self.id_getter(item)
        return params

    def template(self):
        return '\n'.join([
                '<input name="%(tokens_name)s" value="%(tokens_value)s"'
                ' type="hidden" />',
                '<button class="image" type="submit" name="%(name)s"'
                ' title="%(title)s" value="1">',
                '<img src="%(src)s" alt="%(alt)s" />',
                '</button>'
                ])


class RelationshipButtonTableMixin(object):

    button_prefix = ""
    extras_prefix = ""
    button_title = u""
    button_image = ''
    empty_message = _('There are none.')

    ignoreRequest = False
    changed = False

    form_wrapper = InlineViewPageTemplate("""
      <form method="post" tal:attributes="action view/@@absolute_url">
        <tal:block replace="structure view/template" />
        <input type="hidden"
               tal:condition="view/batch"
               tal:attributes="value view/batch/start;
                               name string:${view/extras_prefix}.batch_start" />
        <input type="hidden"
               tal:condition="view/batch"
               tal:attributes="value view/batch/size;
                               name string:${view/extras_prefix}.batch_size" />
      </form>
    """)

    empty_template = InlineViewPageTemplate('''
      <h3 tal:content="view/empty_message" />
    ''')

    def submitItems(self, item):
        raise NotImplementedError

    @property
    def source(self):
        return self.view.getAvailableItemsContainer()

    def columns(self):
        # XXX: evil!
        default = super(RelationshipButtonTableMixin, self).columns()

        action = RelationshipButton(
            self.button_prefix, name='action',
            title=self.button_title, alt=self.button_title,
            library='schooltool.skin.flourish',
            image=self.button_image, id_getter=self.view.getKey)
        return default + [action]

    def update(self):
        super(RelationshipButtonTableMixin, self).update()
        self.applyChanges()

    def applyChanges(self):
        if not self.fromPublication:
            return
        item_prefix = self.button_prefix+'.'
        item_submitted = False
        for param in self.request.form:
            if param.startswith(item_prefix):
                item_submitted = True
        if item_submitted:
            self.changed = True
            self.submitItems()

    def nextURL(self):
        next_url = absoluteURL(self.view, self.request)
        next_url += '?'+self.extra_url()
        if self.batch:
            batch = self.batch
            if self.extras_prefix+'.batch_start' in self.request:
                next_url += "&start%s=%s" % (
                    batch.name, self.request[self.extras_prefix+'.batch_start'])
                if self.extras_prefix+'.batch_size' in self.request:
                    next_url += "&size%s=%s" % (
                        batch.name, self.request[self.extras_prefix+'.batch_size'])
        return next_url

    def renderTable(self):
        # XXX: evil!
        if (self._table_formatter is None or
            not self._items):
            return self.empty_template()
        return super(RelationshipButtonTableMixin, self).renderTable()

    def render(self, *args, **kw):
        if self.changed:
            next = self.nextURL()
            self.request.response.redirect(next)
            return ''
        # XXX: more evil!
        return super(RelationshipButtonTableMixin, self).render(*args, **kw)


class RelationshipAddTableMixin(RelationshipButtonTableMixin):

    button_prefix = "add_item"
    extras_prefix = "on_add"
    button_title = _('Add')
    button_image = 'add-icon.png'

    def submitItems(self):
        for item in self.view.getAvailableItems():
            key = '%s.%s' % (self.button_prefix, self.view.getKey(item))
            if key in self.request:
                self.view.add(removeSecurityProxy(item))

    def updateFormatter(self):
        ommit = self.view.getOmmitedItems()
        columns = self.columns()
        self.setUp(columns=columns,
                   ommit=ommit,
                   table_formatter=self.table_formatter,
                   batch_size=self.batch_size,
                   prefix=self.__name__,
                   css_classes={'table': 'data relationships-table'})


class RelationshipRemoveTableMixin(RelationshipButtonTableMixin):

    button_prefix = "remove_item"
    extras_prefix = "on_remove"
    button_title = _('Remove')
    button_image = 'remove-icon.png'

    def submitItems(self):
        for item in self.view.getSelectedItems():
            key = '%s.%s' % (self.button_prefix, self.view.getKey(item))
            if key in self.request:
                self.view.remove(removeSecurityProxy(item))

    def updateFormatter(self):
        items = self.view.getSelectedItems()
        columns = self.columns()
        self.setUp(columns=columns,
                   items=items,
                   table_formatter=self.table_formatter,
                   batch_size=self.batch_size,
                   prefix=self.__name__,
                   css_classes={'table': 'data relationships-table'})


class AddAllResultsButton(flourish.viewlet.Viewlet):

    template = InlineViewPageTemplate('''
      <div class="buttons">
        <input id="form-buttons-add" name="ADD_DISPLAYED_RESULTS"
               class="submit-widget button-field button-ok"
               value="Add all displayed" type="submit"
               tal:attributes="name view/button_name;
                               value view/title" />
      </div>
    ''')

    title = _('Add all displayed')
    button_name = 'ADD_DISPLAYED_RESULTS'
    token_key = 'displayed.add_item.tokens'

    def addSearchResults(self):
        if (self.button_name not in self.request or
            self.token_key not in self.request):
            return False
        add_ids = self.request[self.token_key]
        if not isinstance(add_ids, list):
            add_ids = [add_ids]
        changed = False
        relationship_view = self.manager.view
        for item in relationship_view.getAvailableItems():
            if relationship_view.getKey(item) in add_ids:
                relationship_view.add(removeSecurityProxy(item))
                changed = True
        return changed

    def update(self):
        changed = self.addSearchResults()
        if changed:
            self.manager.changed = True

    def render(self, *args, **kw):
        if not self.manager._items:
            return ''
        return self.template(*args, **kw)


class EditRelationships(flourish.page.Page):

    content_template = ViewPageTemplateFile('templates/f_tabled_edit_relationships.pt')

    current_title = None
    available_title = None

    def add(self, item):
        collection = removeSecurityProxy(self.getCollection())
        if item not in collection:
            collection.add(item)

    def remove(self, item):
        collection = removeSecurityProxy(self.getCollection())
        if item in collection:
            collection.remove(item)

    def getCollection(self):
        """Return the backend storage for related objects."""
        raise NotImplementedError("Subclasses should override this method.")

    def getSelectedItems(self):
        """Return a sequence of items that are already selected."""
        return self.getCollection()

    def getAvailableItems(self):
        """Return a sequence of items that can be selected."""
        container = self.getAvailableItemsContainer()
        selected_items = set(self.getSelectedItems())
        return [p for p in container.values()
                if p not in selected_items]

    def getAvailableItemsContainer(self):
        """Returns the backend storage for available items."""
        raise NotImplementedError("Subclasses should override this method.")

    def getOmmitedItems(self):
        return self.getSelectedItems()

    def getKey(self, item):
        return item.__name__

    def nextURL(self):
        url = self.request.get('nexturl')
        if url is None:
            url = absoluteURL(self.context, self.request)
        return url


class FlourishRelationshipViewBase(flourish.page.Page):

    content_template = ViewPageTemplateFile('templates/f_edit_relationships.pt')

    current_title = None
    available_title = None

    def add(self, item):
        collection = removeSecurityProxy(self.getCollection())
        if item not in collection:
            collection.add(item)

    def remove(self, item):
        collection = removeSecurityProxy(self.getCollection())
        if item in collection:
            collection.remove(item)

    def getCollection(self):
        """Return the backend storage for related objects."""
        raise NotImplementedError("Subclasses should override this method.")

    def getSelectedItems(self):
        """Return a sequence of items that are already selected."""
        return self.getCollection()

    def getAvailableItems(self):
        """Return a sequence of items that can be selected."""
        container = self.getAvailableItemsContainer()
        selected_items = set(self.getSelectedItems())
        return [p for p in container.values()
                if p not in selected_items]

    def nextURL(self):
        url = self.request.get('nexturl')
        if url is None:
            url = absoluteURL(self.context, self.request)
        return url

    def getAvailableItemsContainer(self):
        """Returns the backend storage for available items."""
        raise NotImplementedError("Subclasses should override this method.")

    def getColumnsAfter(self, prefix):
        actions = {
            'add_item': {'label': _('Add'), 'icon': 'add-icon.png'},
            'remove_item': {'label': _('Remove'), 'icon': 'remove-icon.png'},
            }
        label, icon = actions[prefix]['label'], actions[prefix]['icon']
        action = RelationshipButton(
            prefix, name='action', title=label, alt=label,
            library='schooltool.skin.flourish',
            image=icon, id_getter=self.getKey)
        return [action]

    def createTableFormatter(self, **kwargs):
        container = self.getAvailableItemsContainer()
        formatter = getMultiAdapter((container, self.request),
                                    ITableFormatter)
        formatter.setUp(columns_after=self.getColumnsAfter(kwargs['prefix']),
                        table_formatter=CSSFormatter,
                        **kwargs)
        return formatter

    def getOmmitedItems(self):
        return self.getSelectedItems()

    def setUpTables(self):
        self.available_table = self.createTableFormatter(
            ommit=self.getOmmitedItems(),
            prefix="add_item")

        self.selected_table = self.createTableFormatter(
            filter=lambda l: l,
            items=self.getSelectedItems(),
            prefix="remove_item",
            batch_size=0)

    def getKey(self, item):
        return item.__name__

    def applyFormChanges(self):
        changed = False
        add_item_prefix = 'add_item.'
        remove_item_prefix = 'remove_item.'
        add_item_submitted = False
        remove_item_submitted = False
        for param in self.request.form:
            if param.startswith(add_item_prefix):
                add_item_submitted = True
            elif param.startswith(remove_item_prefix):
                remove_item_submitted = True
        if add_item_submitted:
            for item in self.getAvailableItems():
                if add_item_prefix + self.getKey(item) in self.request:
                    self.add(removeSecurityProxy(item))
                    changed = 'added'
        if remove_item_submitted:
            for item in self.getSelectedItems():
                if remove_item_prefix + self.getKey(item) in self.request:
                    self.remove(removeSecurityProxy(item))
                    changed = 'removed'
        return changed

    def addSearchResults(self):
        key = 'displayed.add_item.tokens'
        if key not in self.request:
            return False
        add_ids = self.request[key]
        if not isinstance(add_ids, list):
            add_ids = [add_ids]
        changes = False
        for item in self.getAvailableItems():
            if self.getKey(item) in add_ids:
                self.add(removeSecurityProxy(item))
                changes = 'added'
        return changes

    def update(self):
        changes = self.applyFormChanges()
        self.setUpTables()

        if not changes:
            if 'ADD_DISPLAYED_RESULTS' in self.request:
                changes = self.addSearchResults()

        if changes:
            this_url = '%s?nexturl=%s' % (str(self.request.URL),
                                          urllib.quote(self.nextURL()))
            # XXX: this be evil hacks indeed
            if changes == 'added':
                this_url += self.available_table.extra_url()
                if self.available_table.batch:
                    batch = self.available_table.batch
                    if 'on_add.batch_start' in self.request:
                        this_url += "&batch_start%s=%s" % (
                            batch.name, self.request['on_add.batch_start'])
                    if 'on_add.batch_size' in self.request:
                        this_url += "&batch_size%s=%s" % (
                            batch.name, self.request['on_add.batch_size'])
            elif changes == 'removed':
                this_url += self.selected_table.extra_url()
                if self.selected_table.batch:
                    batch = self.selected_table.batch
                    if 'on_remove.batch_start' in self.request:
                        this_url += "&batch_start%s=%s" % (
                            batch.name, self.request['on_remove.batch_start'])
                    if 'on_remove.batch_size' in self.request:
                        this_url += "&batch_size%s=%s" % (
                            batch.name, self.request['on_remove.batch_size'])
            self.request.response.redirect(this_url)
            return


class ApplicationLoginView(BrowserView):
    """Backwards compatible login view that redirects to the actual login view."""

    def __call__(self):
        nexturl = absoluteURL(self.context,
                                   self.request) + '/auth/@@login.html'
        if 'nexturl' in self.request:
            nexturl += '?nexturl=' + self.request['nexturl']
        self.request.response.redirect(nexturl)


class ApplicationLogoutView(BrowserView):
    """Backwards compatible logout view that redirects to the actual logout view."""

    def __call__(self):
        nexturl = absoluteURL(self.context,
                              self.request) + '/auth/@@logout.html'
        self.request.response.redirect(nexturl)


class LoginView(BrowserView):
    """A login view"""

    error = None

    def __call__(self):
        self.update()
        return self.index()

    def update(self):
        if ('LOGIN' in self.request and 'username' in self.request and
            'password' in self.request):
            auth = getUtility(IAuthentication)
            try:
                auth.setCredentials(self.request, self.request['username'],
                                    self.request['password'])
            except ValueError:
                self.error = _("Username or password is incorrect")
            else:
                principal = auth.authenticate(self.request)
                person = IPerson(principal, None)
                if 'nexturl' in self.request:
                    nexturl = self.request['nexturl']
                elif person is not None:
                    nexturl = absoluteURL(
                        person, self.request) + '/@@logindispatch'
                else:
                    nexturl = absoluteURL(ISchoolToolApplication(None),
                                               self.request)
                self.request.response.redirect(nexturl)


class LoginDispatchView(BrowserView):
    """Redirects to the proper starting page after login.

    By default the schooltool redirects to the persons calendar.
    """

    def __call__(self):
        nexturl = absoluteURL(
            ISchoolToolCalendar(self.context), self.request)
        self.request.response.redirect(nexturl)


class FlourishLoginDispatchView(BrowserView):
    implements(flourish.interfaces.IPage)

    title = None
    subtitle = None
    has_header = True
    page_class = ''
    container_class = ''
    template = None
    page_template = None
    content_template = None

    def update(self):
        app = ISchoolToolApplication(None)
        manager = queryMultiAdapter(
            (app, self.request, self),
            flourish.interfaces.IContentProvider, 'header_navigation')
        manager.collect()
        apptabs = removeSecurityProxy(IApplicationTabs(app))
        viewlet = manager.get(apptabs.default, None)
        if viewlet is not None and viewlet.enabled:
            nexturl = viewlet.url
        else:
            nexturl = absoluteURL(
                ISchoolToolCalendar(self.context), self.request)
        self.request.response.redirect(nexturl)

    def render(self, *args, **kw):
        return ''

    def __call__(self, *args, **kw):
        self.update()
        if self.request.response.getStatus() in [300, 301, 302, 303,
                                                 304, 305, 307]:
            return u''
        result = self.render(*args, **kw)
        return result


class LogoutView(BrowserView):
    """Clears the authentication creds from the session"""

    def __call__(self):
        auth = getUtility(IAuthentication)
        auth.clearCredentials(self.request)
        url = absoluteURL(ISchoolToolApplication(None),
                               self.request)
        self.request.response.redirect(url)


class LoginNavigationViewlet(flourish.page.LinkViewlet):

    @property
    def authenticated_person(self):
        principal = getattr(self.request, 'principal', None)
        if principal is None:
            return None
        if IUnauthenticatedPrincipal.providedBy(principal):
            return None
        return IPerson(principal, None)

    @property
    def title(self):
        person = self.authenticated_person
        if person is None:
            return _("Log in")
        return _("Log out")

    @property
    def login_url(self):
        app_url = absoluteURL(ISchoolToolApplication(None), self.request)
        return '%s/%s' % (app_url, 'login.html')

    @property
    def logout_url(self):
        app_url = absoluteURL(ISchoolToolApplication(None), self.request)
        return '%s/%s' % (app_url, 'logout.html')

    @property
    def url(self):
        person = self.authenticated_person
        if person is None:
            return self.login_url
        return self.logout_url


class LoginRedirectBackNavigationViewlet(LoginNavigationViewlet):

    @property
    def login_url(self):
        app_url = absoluteURL(ISchoolToolApplication(None), self.request)
        next_url = urllib.quote(str(self.request.URL))
        return '%s/%s?nexturl=%s' % (app_url, 'login.html', next_url)


class CalendarLoginNavigationViewlet(LoginNavigationViewlet):

    @property
    def login_url(self):
        app_url = absoluteURL(ISchoolToolApplication(None), self.request)
        if ISchoolToolApplication.providedBy(self.context.__parent__):
            return '%s/%s' % (app_url, 'login.html')
        next_url = urllib.quote(str(self.request.URL))
        return '%s/%s?nexturl=%s' % (app_url, 'login.html', next_url)


class LoggedInNameViewlet(LoginNavigationViewlet):

    @property
    def url(self):
        person = self.authenticated_person
        if not person:
            return None
        return absoluteURL(person, self.request)

    @property
    def title(self):
        person = self.authenticated_person
        if not person:
            return None
        title_content = queryMultiAdapter(
            (person, self.request, self.view),
            flourish.interfaces.IContentProvider,
            'title')
        if title_content is None:
            return ''
        return title_content.title


class BreadcrumbViewlet(flourish.viewlet.Viewlet):

    def render(self, *args, **kw):
        breadcrumbs = queryMultiAdapter(
            (self.context, self.request, self.view),
            flourish.interfaces.IContentProvider,
            'breadcrumbs')
        if breadcrumbs is None:
            return ''
        return breadcrumbs()


class ApplicationPreferencesView(BrowserView):
    """View used for editing application preferences."""

    __used_for__ = IApplicationPreferences

    error = None
    message = None

    schema = IApplicationPreferences

    def __init__(self, context, request):
        BrowserView.__init__(self, context, request)

        app = ISchoolToolApplication(None)
        prefs = self.schema(app)
        initial = {}
        for field in self.schema:
            initial[field] = getattr(prefs, field)
        setUpWidgets(self, self.schema, IInputWidget, initial=initial)

    def update(self):
        if 'CANCEL' in self.request:
            url = absoluteURL(self.context, self.request)
            self.request.response.redirect(url)
        elif 'UPDATE_SUBMIT' in self.request:
            try:
                data = getWidgetsData(self, self.schema)
            except WidgetsError:
                return # Errors will be displayed next to widgets

            app = ISchoolToolApplication(None)
            prefs = self.schema(app)
            for field in self.schema:
                if field in data: # skip non-fields
                    setattr(prefs, field, data[field])


class FlourishApplicationPreferencesView(Form, form.EditForm):

    fields = field.Fields(IApplicationPreferences)
    fields = fields.select('frontPageCalendar',
                           'timezone',
                           'timeformat',
                           'dateformat',
                           'weekstart',)
    label = None
    legend = _('Calendar settings')

    def getContent(self):
        return IApplicationPreferences(self.context)

    def update(self):
        form.EditForm.update(self)

    @button.buttonAndHandler(_('Submit'), name='apply')
    def handle_edit_action(self, action):
        super(FlourishApplicationPreferencesView, self).handleApply.func(self, action)
        # XXX: hacky sucessful submit check
        if (self.status == self.successMessage or
            self.status == self.noChangesMessage):
            self.request.response.redirect(self.nextURL())

    @button.buttonAndHandler(_('Cancel'))
    def handle_cancel_action(self, action):
        self.request.response.redirect(self.nextURL())

    def nextURL(self):
        url = absoluteURL(self.context, self.request) + '/settings'
        return url

    def updateActions(self):
        super(FlourishApplicationPreferencesView, self).updateActions()
        self.actions['apply'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')


class FlourishSchoolNameEditView(FlourishApplicationPreferencesView):

    fields = field.Fields(IApplicationPreferences).select('title', 'logo')
    legend = _('School Name')

    def updateActions(self):
        super(FlourishSchoolNameEditView, self).updateActions()
        self.actions['apply'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    def updateWidgets(self):
        super(FlourishSchoolNameEditView, self).updateWidgets()
        self.widgets['title'].label = _('Name')
        self.widgets['logo'].label = _('Logo')

    def nextURL(self):
        url = absoluteURL(self.context, self.request) + '/manage'
        return url


class ProbeParticipation:
    """A stub participation for use in hasPermissions."""
    implements(IParticipation)
    def __init__(self, principal):
        self.principal = principal
        self.interaction = None


def hasPermissions(permissions, object, principalid):
    """Test if the principal has access according to the security policy."""
    principal = getUtility(IAuthentication).getPrincipal(principalid)
    participation = ProbeParticipation(principal)
    interaction = getSecurityPolicy()(participation)
    return [interaction.checkPermission(permission, object)
            for permission in permissions]


class LeaderView(RelationshipViewBase):

    __used_for__ = IAsset

    title = _("Leaders")
    current_title = _("Current leaders")
    available_title = _("Available leaders")

    def getCollection(self):
        return self.context.leaders

    def getAvailableItemsContainer(self):
        return ISchoolToolApplication(None)['persons']


class FlourishLeaderView(EditRelationships):

    current_title = _("Current responsible parties")
    available_title = _("Available responsible parties")

    def getCollection(self):
        return self.context.leaders

    def getAvailableItemsContainer(self):
        return ISchoolToolApplication(None)['persons']


class ManageView(BrowserView):
    pass


class ManageMenuViewletManager(OrderedViewletManager):
    implements(IManageMenuViewletManager)


class ViewRobot(BrowserPage):

    def __call__(self):
        return "User-agent: *\nDisallow: /"


@adapter(ISchoolToolApplication)
@implementer(ISchoolToolAuthenticationPlugin)
def getAuthentication(app):
    return LocationProxy(getUtility(ISchoolToolAuthenticationPlugin),
                         app, 'auth')


def getCharset(content_type, default="UTF-8"):
    """Get charset out of content-type

        >>> getCharset('text/xml; charset=latin-1')
        'latin-1'

        >>> getCharset('text/xml; charset=yada-yada')
        'yada-yada'

        >>> getCharset('text/xml; charset=yada-yada; fo=ba')
        'yada-yada'

        >>> getCharset('text/plain')
        'UTF-8'

        >>> getCharset(None)
        'UTF-8'

    """
    if not content_type:
        return default

    parts = content_type.split(";")
    if len(parts) == 0:
        return default

    stripped_parts = [part.strip() for part in parts]

    charsets = [part for part in stripped_parts
                if part.startswith("charset=")]

    if len(charsets) == 0:
        return default

    return charsets[0].split("=")[1]


class HTTPCalendarView(BrowserView):
    """Restive view for calendars"""

    def GET(self):
        data = "\r\n".join(convert_calendar_to_ical(self.context)) + "\r\n"
        request = self.request
        request.response.setHeader('Content-Type',
                                   'text/calendar; charset=UTF-8')
        request.response.setHeader('Content-Length', len(data))
        request.response.setStatus(200)
        return data

    def PUT(self):
        request = self.request

        for name in request:
            if name.startswith('HTTP_CONTENT_'):
                # Unimplemented content header
                request.response.setStatus(501)
                return ''

        body = self.request.bodyStream
        data = body.read()
        charset = getCharset(self.request.getHeader("Content-Type"))

        from schooltool.app.interfaces import IWriteCalendar
        adapter = IWriteCalendar(self.context)
        adapter.write(data, charset)
        return ''


class TitleView(BrowserView):

    def __call__(self):
        return self.context.title


class ContentTitle(ContentProvider):
    render = InlineViewPageTemplate('''
        <span tal:replace="structure view/title"></span>
    '''.strip())

    @property
    def title(self):
        __not_set = object()
        title = getattr(self.context, 'title', __not_set)
        if title is __not_set:
            return getattr(self.context, '__name__', None)
        return title


class ContentLink(ContentTitle):
    render = InlineViewPageTemplate('''
        <a tal:attributes="href view/url" tal:content="view/title"></a>
    '''.strip())

    def url(self):
        return absoluteURL(self.context, self.request)


class ContentLabel(ContentLink):
    def title(self):
        __not_set = object()
        label = getattr(self.context, 'label', __not_set)
        if label is __not_set:
            return super(ContentLabel, self).title
        return label


class ManageSiteBreadcrumb(flourish.breadcrumbs.Breadcrumbs):

    follow_crumb = None
    title = _('Server')

    @property
    def url(self):
        app = ISchoolToolApplication(None)
        app_url = absoluteURL(app, self.request)
        return '%s/settings' % app_url


class ManageSite(flourish.page.Page):
    pass


class ServerActionsLinks(flourish.page.RefineLinksViewlet):
    """Server actions links viewlet."""


class ManageSchool(flourish.page.Page):
    pass


class ActiveSchoolYearContentMixin(object):

    @Lazy
    def schoolyear(self):
        schoolyears = ISchoolYearContainer(ISchoolToolApplication(None))
        result = schoolyears.getActiveSchoolYear()
        if 'schoolyear_id' in self.request:
            schoolyear_id = self.request['schoolyear_id']
            result = schoolyears.get(schoolyear_id, result)
        return result

    @property
    def has_schoolyear(self):
        return self.schoolyear is not None

    def url_with_schoolyear_id(self, obj, view_name=''):
        schoolyear = self.schoolyear
        schoolyear_id = schoolyear.__name__ if schoolyear is not None else ''
        params = {'schoolyear_id': schoolyear_id.encode('utf-8')}
        context_url = absoluteURL(obj, self.request)
        return '%s/%s?%s' % (context_url, view_name, urllib.urlencode(params))


class ManageItemDoneLink(flourish.viewlet.Viewlet,
                         ActiveSchoolYearContentMixin):
    template = InlineViewPageTemplate('''
      <h3 tal:define="can_manage view/can_manage"
          class="done-link"
          i18n:domain="schooltool">
        <tal:block condition="can_manage">
          <a tal:attributes="href view/manage_url"
             i18n:translate="">Done</a>
        </tal:block>
        <tal:block condition="not:can_manage">
          <a tal:attributes="href request/principal/schooltool:person/@@absolute_url"
             i18n:translate="">Done</a>
        </tal:block>
      </h3>
      ''')

    @property
    def can_manage(self):
        return flourish.canEdit(self.context) or \
               inCrowd(self.request.principal, 'administration', self.context)

    def manage_url(self):
        app = ISchoolToolApplication(None)
        return self.url_with_schoolyear_id(app, view_name='manage')


class ManageSiteLinks(flourish.page.RefineLinksViewlet):
    """Manage Site links viewlet."""


class CustomizeSchoolLinks(flourish.page.RefineLinksViewlet):
    """Customize School links viewlet."""


class SchoolAddLinks(flourish.page.RefineLinksViewlet):
    """School add links viewlet."""


class SchoolActionsLinks(flourish.page.RefineLinksViewlet):
    """School actions links viewlet."""


class AboutLinks(flourish.page.RefineLinksViewlet):
    """About links viewlet."""


def getAppViewlet(context, request, view, manager, name):
    app = ISchoolToolApplication(None)
    viewlet = flourish.viewlet.lookupViewlet(
        app, request, view, manager, name=name)
    return viewlet


class ApplicationControlLinks(flourish.page.RefineLinksViewlet):
    """Application Control links viewlet."""


class FlourishRuntimeInfoView(RuntimeInfoView, ZODBControlView):

    def dbSettings(self):
        result = {}
        database_settings = self.databases[0]
        result['dbSize'] = database_settings['size']
        result['dbName'] = database_settings['dbName']
        return result

    def update(self):
        pass


class FlourishServerSettingsOverview(flourish.page.Content,
                                     ZODBControlView,
                                     RuntimeInfoView):

    body_template = ViewPageTemplateFile(
        'templates/f_server_settings_overview.pt')

    @property
    def settings(self):
        result = {}
        database_settings = self.databases[0]
        result['dbSize'] = database_settings['size']
        result['dbName'] = database_settings['dbName']
        control_settings = (
            'Uptime',
            'SystemPlatform',
            'PythonVersion',
            'CommandLine',
            )
        runtimeInfo = self.runtimeInfo()
        for setting in control_settings:
            result[setting] = runtimeInfo.get(setting)
        return result

    def runtimeInfo(self):
        application_control = IApplicationControl(self.context)
        try:
            ri = IRuntimeInfo(application_control)
        except TypeError:
            formatted = dict.fromkeys(self._fields, self._unavailable)
            formatted["Uptime"] = self._unavailable
        else:
            formatted = self._getInfo(ri)
        return formatted


class PackageVersionsOverview(flourish.page.Content):

    body_template = ViewPageTemplateFile(
        'templates/package_versions.pt')


class FlourishCalendarSettingsOverview(flourish.form.FormViewlet):

    fields = field.Fields(IApplicationPreferences)
    fields = fields.select('frontPageCalendar', 'timezone', 'timeformat',
                           'dateformat', 'weekstart')
    template = ViewPageTemplateFile(
        'templates/f_calendar_settings_overview.pt')
    mode = DISPLAY_MODE


class PackDatabaseLink(flourish.page.ModalFormLinkViewlet):

    @property
    def dialog_title(self):
        title = _(u'Pack Database')
        return translate(title, context=self.request)


class PackDatabaseView(Dialog):

    def initDialog(self):
        super(PackDatabaseView, self).initDialog()
        self.ajax_settings['dialog']['dialogClass'] = 'explicit-close-dialog'
        self.ajax_settings['dialog']['closeOnEscape'] = False

    def update(self):
        Dialog.update(self)
        if 'DONE' in self.request:
            url = absoluteURL(self.context, self.request) + '/settings'
            self.request.response.redirect(url)
            return
        days = 0
        db = getUtility(IDatabase, name='')
        try:
            db.pack(days=days)
            self.status = _('Database successfully packed.')
        except FileStorageError, err:
            self.status = _('There were errors while packing the database: ${error}',
                            mapping={'error': err})


class FlourishErrorsViewBase(flourish.page.Page):

    @property
    def error_utility(self):
        default_site = self.context.getSiteManager().get('default')
        return getUtility(IErrorReportingUtility, context=default_site)


class FlourishErrorsView(FlourishErrorsViewBase):

    def formatEntryValue(self, value):
        return len(value) < 70 and value or value[:70] + '...'

    def getLogEntries(self):
        return self.error_utility.getLogEntries()

    def settings(self):
        result = {}
        properties = self.error_utility.getProperties()
        for setting in ('keep_entries', 'ignored_exceptions'):
            result[setting] = properties[setting]
        if properties['copy_to_zlog']:
            result['copy_to_zlog'] = _('Yes')
        else:
            result['copy_to_zlog'] = _('No')
        return result


class FlourishErrorEntryView(FlourishErrorsViewBase):

    def update(self):
        entryId = self.request.get('id')
        if not entryId:
            self.request.response.redirect(self.nextURL())
            return
        self.entry = self.error_utility.getLogEntryById(entryId)
        if self.entry is not None:
            error_type = self.entry['type']
            self.subtitle = _('${type} Error', mapping={'type': error_type})

    def nextURL(self):
        return absoluteURL(self.context, self.request) + '/errors'


class IErrorsSettings(Interface):

    keep_entries = Int(
        title=_('Number of exceptions to show'))

    copy_to_zlog = Bool(
        title=_('Copy exceptions to the event log'),
        default=False)

    ignored_exceptions = Tuple(
        title=_('Ignored exception types'),
        value_type=Choice(vocabulary=vocabulary([
                    ('Unauthorized', 'Unauthorized'),
                    ('NotFound', 'NotFound')])),
        required=False)


class ErrorsSettingsAdapter(object):

    adapts(IErrorReportingUtility)
    implements(IErrorsSettings)

    def __init__(self, context):
        self.__dict__['context'] = context

    def __getattr__(self, name):
        if name == 'ignored_exceptions':
            return getattr(self.context, '_ignored_exceptions')
        return getattr(self.context, name)

    def __setattr__(self, name, value):
        if name == 'ignored_exceptions':
            self.context._ignored_exceptions = value
            return
        setattr(self.context, name, value)


class FlourishErrorsConfigureView(Form, FlourishErrorsViewBase):

    legend = _('Errors Settings')
    fields = field.Fields(IErrorsSettings)
    fields['ignored_exceptions'].widgetFactory = CheckBoxFieldWidget

    def getContent(self):
        return self.error_utility

    @button.buttonAndHandler(_('Submit'))
    def handle_submit_action(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return
        form.applyChanges(self, self.getContent(), data)
        self.request.response.redirect(self.nextURL())

    @button.buttonAndHandler(_('Cancel'))
    def handle_cancel_action(self, action):
        self.request.response.redirect(self.nextURL())

    def nextURL(self):
        url = absoluteURL(self.context, self.request) + '/errors'
        return url

    def updateActions(self):
        super(FlourishErrorsConfigureView, self).updateActions()
        self.actions['submit'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')


class ErrorsBreadcrumb(flourish.breadcrumbs.Breadcrumbs):

    title = _('Errors')

    @property
    def url(self):
        app = ISchoolToolApplication(None)
        app_url = absoluteURL(app, self.request)
        return '%s/errors' % app_url

    @property
    def follow_crumb(self):
        return ManageSiteBreadcrumb(self.context, self.request, self.view)


class FlourishHideUnhideTabsView(flourish.page.Page):

    ignore_tabs = ['manage_school', 'manage_site']

    @property
    def tabs(self):
        tabs = []
        apptabs = removeSecurityProxy(IApplicationTabs(self.context))
        manager = queryMultiAdapter(
            (self.context, self.request, self),
            flourish.interfaces.IContentProvider, 'header_navigation')
        manager.collect()
        for name in manager.order:
            if name in self.ignore_tabs:
                continue
            tabs.append({
                'title': removeSecurityProxy(manager[name]).title,
                'name': name,
                'checked': apptabs.get(name, True) and 'checked' or '',
                'default': name == apptabs.default and 'checked' or '',
                })
        return tabs

    def update(self):
        if 'CANCEL' in self.request:
            self.request.response.redirect(self.nextURL())
        elif 'SUBMIT' in self.request:
            default = self.request.get('default_tab')
            visible = set(self.request.get('visible', []))
            visible.add(default)
            names = set([tab['name'] for tab in self.tabs])
            if default in names and visible.issubset(names):
                apptabs = removeSecurityProxy(IApplicationTabs(self.context))
                apptabs.default = default
                for name in names:
                    if name in visible:
                        if not apptabs.get(name, True):
                            apptabs[name] = True
                    else:
                        apptabs[name] = False
            self.request.response.redirect(self.nextURL())

    def nextURL(self):
        return absoluteURL(self.context, self.request) + '/settings'


class TabsBreadcrumb(flourish.breadcrumbs.Breadcrumbs):

    title = _('Tabs')

    @property
    def url(self):
        app = ISchoolToolApplication(None)
        app_url = absoluteURL(app, self.request)
        return '%s/hide_unhide_tabs.html' % app_url

    @property
    def follow_crumb(self):
        return ManageSiteBreadcrumb(self.context, self.request, self.view)


class FlourishAboutView(flourish.page.Page):

    pass


class SchoolLogoView(flourish.widgets.ImageView):

    @property
    def image(self):
        app = ISchoolToolApplication(None)
        prefs = IApplicationPreferences(app)
        return prefs.logo


class SchoolLogoViewlet(flourish.viewlet.Viewlet):

    template = InlineViewPageTemplate("""
    <div class="header">
      <div class="photo-display">
        <img tal:attributes="src view/url; alt view/title" />
      </div>
    </div>
    """)

    @property
    def enabled(self):
        app = ISchoolToolApplication(None)
        prefs = IApplicationPreferences(app)
        return prefs.logo is not None

    @property
    def url(self):
        app = ISchoolToolApplication(None)
        base = absoluteURL(app, self.request)
        return '%s/logo' % base

    def render(self, *args, **kw):
        if not self.enabled:
            return ""
        return self.template(*args, **kw)


class SchoolLoginLogoViewlet(SchoolLogoViewlet):

    template = InlineViewPageTemplate("""
    <div class="header">
      <div class="photo-display">
        <img tal:attributes="src view/url; alt view/title" />
      </div>
    </div>
    <div class="body">
      <div>
        <h3 tal:content="context/schooltool:app/title" />
      </div>
    </div>
    """)


class ManageSchoolViewlet(flourish.page.LinkViewlet):

    @property
    def enabled(self):
        if not self.title:
            return False
        return inCrowd(self.request.principal, 'administration', context=self.context)


class NameSortingEditView(FlourishApplicationPreferencesView):

    fields = field.Fields(IApplicationPreferences).select('name_sorting')
    legend = _('Name Sorting')

    def updateActions(self):
        super(NameSortingEditView, self).updateActions()
        self.actions['apply'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    def updateWidgets(self):
        super(NameSortingEditView, self).updateWidgets()
        self.widgets['name_sorting'].label = _('Sort by')

    def nextURL(self):
        url = absoluteURL(self.context, self.request) + '/settings'
        return url


class NameSortingBreadcrumb(flourish.breadcrumbs.Breadcrumbs):

    title = _('Name Sorting')

    @property
    def url(self):
        app = ISchoolToolApplication(None)
        app_url = absoluteURL(app, self.request)
        return '%s/name_sorting.html' % app_url

    @property
    def follow_crumb(self):
        return ManageSiteBreadcrumb(self.context, self.request, self.view)
